import unittest
import numpy as np

from inmoose.utils import Factor
from inmoose.deseq2 import DESeq, makeExampleDESeqDataSet


class Test(unittest.TestCase):
    def test_results(self):
        """test that results work as expected and throw errors"""
        ## test contrasts
        dds = makeExampleDESeqDataSet(n=200, m=12, seed=42)
        dds.obs["condition"] = Factor(np.repeat([1, 2, 3], 4))
        dds.obs["group"] = Factor(np.repeat([[1, 2]], 6, axis=0).flatten())
        dds.obs["foo"] = np.repeat(["lo", "hi"], 6)
        dds.counts()[:, 0] = np.repeat([100, 200, 800], 4)

        dds.design = "~ group + condition"

        # calling results too early
        with self.assertRaisesRegexp(
            ValueError,
            expected_regex="could not find results in obj. first run DESeq()",
        ):
            dds.results()

        dds.sizeFactors = np.ones(dds.n_obs)
        dds = DESeq(dds)
        res = dds.results()
        # TODO
        # show_res = res.show()
        summary = res.summary()
        print(summary)
        summary_ref = """
out of 200 with nonzero total read count
adjusted p-value < 0.1
LFC > 0 (up)       : 1, 0.50%
LFC < 0 (down)     : 0, 0.00%
outliers [1]       : 0, 0.00%
low counts [2]     : 0, 0.00%
(mean count < 0)
[1] see 'cooksCutoff' argument of results()
[2] see 'independentFiltering' argument of results()
"""
        self.assertEqual(summary, summary_ref)

        # various results error checking
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="the LRT requires the user to run nbinomLRT or DESeq",
        ):
            dds.results(test="LRT")
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="when testing altHypothesis='lessAbs', set the argument lfcThreshold to a positive value",
        ):
            dds.results(altHypothesis="lessAbs")
        with self.assertRaisesRegex(
            ValueError, expected_regex="'name' should be a string"
        ):
            dds.results(name=["Intercept", "group1"])
        with self.assertRaisesRegex(ValueError, expected_regex="foo is not a factor"):
            dds.results(contrast=["foo", "B", "A"])
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="as 1 is the reference level, was expecting condition_4_vs_1 to be present in",
        ):
            dds.results(contrast=["condition", "4", "1"])
        with self.assertRaisesRegex(
            ValueError, expected_regex="invalid value for test: foo"
        ):
            dds.results(test="foo")
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="numeric contrast vector should have one element for every element of",
        ):
            dds.results(contrast=False)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="'contrast', as a pair of lists, should have length 2",
        ):
            dds.results(contrast=["a", "b", "c", "d"])
        with self.assertRaisesRegex(
            ValueError, expected_regex="1 and 1 should be different level names"
        ):
            dds.results(contrast=["condition", "1", "1"])

        dds.results(independentFiltering=False)
        dds.results(contrast=["condition_2_vs_1"])

        with self.assertRaisesRegex(
            ValueError,
            expected_regex="condition_3_vs_1 and condition_3_vs_1 should be different level names",
        ):
            dds.results(
                contrast=["condition_2_vs_1", "condition_3_vs_1", "condition_3_vs_1"]
            )
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="'contrast', as a pair of lists, should have lists of strings as elements",
        ):
            dds.results(contrast=["condition_2_vs_1", 1])
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="all elements of the 2-element contrast should be elements of",
        ):
            dds.results(contrast=["condition_2_vs_1", "foo"])
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="elements in the 2-element contrast should only appear in the numerator",
        ):
            dds.results(contrast=["condition_2_vs_1", "condition_2_vs_1"])
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="all elements of the 2-element contrast should be elements of",
        ):
            dds.results(contrast=["", ""])
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="numeric contrast vector should have one element for every element of",
        ):
            dds.results(contrast=np.repeat(0, 6))
        with self.assertRaisesRegex(ValueError, expected_regex="foo is not a factor"):
            dds.results(contrast=["foo", "lo", "hi"])

        self.assertAlmostEqual(
            dds.results(contrast=["condition", "1", "3"]).log2FoldChange[0],
            -3,
            delta=1e-6,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "1", "2"]).log2FoldChange[0],
            -1,
            delta=1e-6,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "2", "3"]).log2FoldChange[0],
            -2,
            delta=1e-6,
        )

        # test a number of contrast as list options
        self.assertAlmostEqual(
            dds.results(
                contrast=["condition_3_vs_1", "condition_2_vs_1"]
            ).log2FoldChange[0],
            2,
            delta=1e-6,
        )
        dds.results(
            contrast=["condition_3_vs_1", "condition_2_vs_1"], listValues=[0.5, -0.5]
        )
        dds.results(contrast=["condition_3_vs_1", []])
        dds.results(contrast=["condition_3_vs_1", []], listValues=[0.5, -0.5])
        dds.results(contrast=[[], "condition_2_vs_1"])
        dds.results(contrast=[[], "condition_2_vs_1"], listValues=[0.5, -0.5])

        # test no prior on intercept
        self.assertTrue(np.array_equal(dds.betaPriorVar, np.repeat(1e6, 4)))

        # test thresholding
        dds.results(lfcThreshold=np.log2(1.5))
        dds.results(lfcThreshold=1, altHypothesis="lessAbs")
        dds.results(lfcThreshold=1, altHypothesis="greater")
        dds.results(lfcThreshold=1, altHypothesis="less")

        dds3 = DESeq(dds, betaPrior=True)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="testing altHypothesis='lessAbs' requires setting the DESeq\(\) argument betaPrior=False",
        ):
            dds3.results(lfcThreshold=1, altHypothesis="lessAbs")

    def test_results_zero_intercept(self):
        """test results on designs with zero intercept"""
        dds = makeExampleDESeqDataSet(n=100, m=12, seed=42)
        dds.obs["condition"] = Factor(np.repeat([1, 2, 3], 4))
        dds.obs["group"] = Factor(np.repeat([[1, 2]], 6, axis=0).flatten())

        dds.X[:, 0] = np.repeat([100, 200, 400], 4)

        dds.design = "~ 0 + condition"
        dds = DESeq(dds, betaPrior=False)

        self.assertAlmostEqual(dds.results().log2FoldChange[0], 2, delta=0.1)
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "2", "1"]).log2FoldChange[0],
            1.25,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "3", "2"]).log2FoldChange[0],
            0.68,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "1", "3"]).log2FoldChange[0],
            -2,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "1", "2"]).log2FoldChange[0],
            -1.25,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "2", "3"]).log2FoldChange[0],
            -0.68,
            delta=0.1,
        )
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="condition\[4\] and condition\[1\] are expected to be in",
        ):
            dds.results(contrast=["condition", "4", "1"])

        dds.design = "~ 0 + group + condition"
        dds = DESeq(dds, betaPrior=False)

        self.assertAlmostEqual(dds.results().log2FoldChange[0], 2, delta=0.1)
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "3", "1"]).log2FoldChange[0],
            2,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "2", "1"]).log2FoldChange[0],
            1.25,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "3", "2"]).log2FoldChange[0],
            0.68,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "1", "3"]).log2FoldChange[0],
            -2,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "1", "2"]).log2FoldChange[0],
            -1.25,
            delta=0.1,
        )
        self.assertAlmostEqual(
            dds.results(contrast=["condition", "2", "3"]).log2FoldChange[0],
            -0.68,
            delta=0.1,
        )

    @unittest.skip("LRT is not implemented yet")
    def test_results_likelihood_ratio_test(self):
        """test results with likelihood ratio test"""
        dds = makeExampleDESeqDataSet(n=100)
        dds.obs["group"] = Factor([1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2])
        dds.design = "~ group + condition"
        dds = DESeq(dds, test="LRT", reduced="~group")

        self.assertFalse(
            np.all(
                dds.results(name="condition_B_vs_A").stat
                == dds.results(name="condition_B_vs_A", test="Wald").stat
            )
        )

        # LFC are already MLE
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="addMLE=TRUE is only for when a beta prior was used",
        ):
            dds.results(addMLE=True)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="tests of log fold change above or below a theshold must be Wald tests",
        ):
            dds.results(lfcThreshold=1, test="LRT")

        self.assertTrue(
            np.all(
                dds.results(test="LRT", contrast=["group", "1", "2"]).log2FoldChange
                == -dds.results(test="LRT", contrast=["group", "2", "1"]).log2FoldChange
            )
        )

    def test_results_basics(self):
        """test that results basics regarding format, saveCols, tidy, MLE, remove are working"""
        dds = makeExampleDESeqDataSet(n=100)
        dds.var["score"] = np.arange(1, 101)
        dds = DESeq(dds)

        # try saving metadata columns
        res = dds.results(saveCols="score")  # string

        # check tidy-ness (unimplemented)
        with self.assertRaises(NotImplementedError):
            res = dds.results(tidy=True)
            self.assertTrue(res.columns[0] == "rows")

        # test MLE and 'name'
        dds2 = DESeq(dds, betaPrior=True)
        dds2.results(addMLE=True)
        with self.assertRaises(ValueError):
            dds2.results(name="condition_B_vs_A", addMLE=True)

        # test remove results
        dds = dds.removeResults()
        self.assertTrue(dds.var.description.filter("results").empty)

    @unittest.skip("not sure what to test")
    def test_results_custom_filters(self):
        """test that custom filters can be provided to results()"""
        dds = makeExampleDESeqDataSet(n=200, m=4, betaSD=np.repeat([0, 2], [150, 50]))
        dds = DESeq(dds)
        res = dds.results()
        method = "BH"
        alpha = 0.1

        raise NotImplementedError()



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/parallel.R' of the Bioconductor DESeq2
# package (version 3.16).


import numpy as np
import pandas as pd

from .fitNbinomGLMs import fitNbinomGLMs
from .misc import (
    renameModelMatrixColumns,
    buildDataFrameWithNACols,
    buildMatrixWithNACols,
)


def estimateMLEForBetaPriorVar(
    obj, maxit=100, useOptim=True, useQR=True, modelMatrixType=None
):
    """
    TODO
    """
    # this function copies code from other functions,
    # in order to allow parallelization
    objNZ = obj[:, ~obj.var["allZero"]]

    if modelMatrixType is None:
        # this code copied from nbinomWaldTest()
        termsOrder = np.array([len(t.factors) for t in obj.design.design_info.terms])
        blindDesign = np.array_equal(termsOrder, [0])
        mmTypeTest = not blindDesign
        if mmTypeTest:
            modelMatrixType = "expanded"
        else:
            modelMatrixType = "standard"

    obj.modelMatrixType = modelMatrixType

    # this code copied from fitGLMsWithPrior()
    fit = fitNbinomGLMs(
        objNZ,
        maxit=maxit,
        useOptim=useOptim,
        useQR=useQR,
        renameCols=(modelMatrixType == "standard"),
    )
    modelMatrix = fit["modelMatrix"]
    modelMatrixNames = modelMatrix.design_info.column_names
    H = fit["hat_diagonals"]
    betaMatrix = fit["betaMatrix"]

    convertNames = renameModelMatrixColumns(obj.obs, objNZ.design)
    modelMatrixNames = [
        convertNames[x] if x in convertNames else x for x in modelMatrixNames
    ]

    mleBetaMatrix = fit["betaMatrix"]
    mleBetaMatrix.columns = [f"MLE_{n}" for n in modelMatrixNames]
    mleBetaMatrix = buildDataFrameWithNACols(mleBetaMatrix, obj.var["allZero"])
    mleBetaMatrix.index = obj.var_names
    # remove any MLE columns if they exist
    obj.var = obj.var.drop(obj.var.filter(regex="MLE_").columns)
    obj.var = pd.concat([obj.var, mleBetaMatrix], axis=1)
    obj.layers["H"] = buildMatrixWithNACols(H, obj.var["allZero"])
    return obj



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/movingAverageByCol.R' of the Bioconductor edgeR package (version 3.38.4).


import logging
import numpy as np
from math import ceil, floor


def movingAverageByCol(x, width=5, full_length=True):
    """
    Moving average smoother for matrix columns.

    Arguments
    ---------
    x : array_like
        numeric matrix
    width : int
        width of window of rows to be averaged
    full_length : bool
        whether the output should have the same number of rows as the input

    Returns
    -------
    ndarray
        numeric matrix with smoothed values. If :code:`full_length = True`, of
        same shape as :code:`x`. If :code:`full_length = False`, has
        :code:`width-1` fewer rows than :code:`x`.
    """
    x = np.asanyarray(x, order="F")
    width = int(width)
    if width <= 1:
        return x
    (n, m) = x.shape
    if width > n:
        width = n
        logging.warnings.warn("reducing moving average width to x.shape[0]")

    if full_length:
        half1 = ceil(width / 2)
        half2 = floor(width / 2)
        x = np.vstack([np.zeros((half1, m)), x, np.zeros((half2, m))])
    else:
        if width == n:
            return x.mean(axis=0).reshape((1, m))
        x = np.vstack([np.zeros((1, m)), x])

    n2 = x.shape[0]
    x = x.cumsum(axis=0)
    x = x[width:n2] - x[0 : (n2 - width)]
    n3 = x.shape[0]
    w = np.full((n3,), width)
    if full_length:
        if half1 > 1:
            w[0 : half1 - 1] = np.array([width - i for i in range(half1 - 1, 0, -1)])
        w[(n3 - half2) : n3] = np.array([width - i for i in range(1, half2 + 1)])
    return x / w.reshape((n3, 1))



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/dispCoxReidInterpolateTagwise.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from math import floor

from .adjustedProfileLik import adjustedProfileLik
from .aveLogCPM import aveLogCPM
from .makeCompressedMatrix import _compressOffsets, _compressWeights
from .maximizeInterpolant import maximizeInterpolant
from .movingAverageByCol import movingAverageByCol


def dispCoxReidInterpolateTagwise(
    y,
    design,
    dispersion,
    offset=None,
    trend=True,
    AveLogCPM=None,
    min_row_sum=5,
    prior_df=10,
    span=0.3,
    grid_npts=11,
    grid_range=(-6, 6),
    weights=None,
):
    """
    Estimate genewise dispersion parameters across multiple negative binomial
    GLMs using weighted Cox-Reid adjusted profile likelihood and cubic spline
    interpolation over a genewise grid.

    In the context of :code:`edgepy`, :func:`dispCoxReidInterpolateTagwise` is a
    low-level function called by :func:`estimateGLMTagwiseDisp`.

    This function calls :func:`maximizeInterpolant` to fit cubic spline
    interpolation over a genewise grid.

    Note that the terms "tag" and "gene" are synonymous here. The function is
    only named "tagwise" for historical reasons.

    Arguments
    ---------
    y : matrix
        matrix of counts
    design : matrix
        design matrix for the GLM to fit
    dispersion : float or array_like
        scalar or vector giving the dispersion(s) towards which the genewise
        dispersion parametes are shrunk
    offset : float or array_like, optional
        scalar, vector or matrix giving the offset (in addition to the log of
        the effective library size) that is to be included in the NB GLM for the
        genes. If a scalar, then this value is used as an offset for all genes
        and libraries. If a vector, it should have length equal to the number of
        libraries, and the same vector of offsets is used for each gene. If a
        matrix, then each library for each gene has its unique offset. In
        :func:`adjustedProfileLik` the :code:`offset` must be a matrix with the
        same shape as the matrix of counts.
    trend : bool, optional
        whether abundance-dispersion trend is used for smoothing
    AveLogCPM : array_like, optional
        vector of average log2 counts per million for each gene
    min_row_sum : int, optional
        value to filter out low abundance genes. Only genes with total sum of
        counts above this threshold are used. Low abundance genes can adversely
        affect the estimation of the common dispersion, so this argument allows
        the user to select an appropriate filter threshold for gene abundance.
        Defaults to 5.
    prior_df : float, optional
        prior desmoothing parameter that indicates the weight to give to the
        common likelihood compared to the individual gene's likelihood; default
        :code:`getPriorN(obj)` gives a value for :code:`prior_n` that is
        equivalent to giving the common likelihood 20 prior degrees of freedom
        in the estimation of the genewise dispersion.
    span : float, optional
        parameter between 0 and 1 specifying proportion of data to be used in
        the local regression moving window. Larger values give smoother fits.
    grid_npts : int, optional
        the number of points at which to place knots for the spline-based
        estimation of the genewise dispersion estimates.
    grid_range : tuple, optional
        relative range, in terms of :code:`log2(dispersion)`, on either side of
        trendline for each gene for spline grid points.
    weights : matrix, optional
        observation weights

    Returns
    -------
    ndarray
        vector of genewise dispersion, same length as the number of genes in
        the input matrix of counts
    """
    # Check y
    y = np.asarray(y, order="F")
    (ntags, nlibs) = y.shape

    # Check design
    design = np.asarray(design, order="F")
    if np.linalg.matrix_rank(design) != design.shape[1]:
        raise ValueError("design matrix must be full column rank")
    ncoefs = design.shape[1]
    if ncoefs >= nlibs:
        raise ValueError("no residual degrees of freedom")

    # Check offset
    lib_size = None
    if offset is None:
        lib_size = y.sum(axis=0)
        offset = np.log(lib_size)
    if len(offset.shape) == 1:
        offset = np.full(y.shape, offset, order="F")
    assert offset.shape == y.shape

    # Check AveLogCPM
    if AveLogCPM is None:
        AveLogCPM = aveLogCPM(y, lib_size=lib_size)
    AveLogCPM = np.asarray(AveLogCPM, order="F")

    # Check dispersion
    dispersion = np.asanyarray(dispersion, order="F")
    if len(dispersion.shape) == 0:
        dispersion = np.full((ntags,), dispersion, order="F")
    else:
        if len(dispersion) != ntags:
            raise ValueError("length of dispersion does not match nrow(y)")

    # Apply min_row_sum and use input dispersion for small count tags
    i = y.sum(axis=1) >= min_row_sum
    if np.logical_not(i).any():
        if i.any():
            dispersion[i] = dispCoxReidInterpolateTagwise(
                y=y[i, :],
                design=design,
                offset=offset[i, :],
                dispersion=dispersion[i],
                AveLogCPM=AveLogCPM[i],
                grid_npts=grid_npts,
                min_row_sum=0,
                prior_df=prior_df,
                span=span,
                trend=trend,
                weights=(weights[i, :] if weights is not None else None),
            )
        return dispersion

    # Posterior profile likelihood
    prior_n = prior_df / (nlibs - ncoefs)
    spline_pts = [
        grid_range[0] + i * (grid_range[1] - grid_range[0]) / (grid_npts - 1)
        for i in range(grid_npts)
    ]
    apl = np.zeros((ntags, grid_npts), order="F")
    # anticipate the calls to _compress* in adjustedProfileLik
    y = np.asarray(y, order="F")
    offset = _compressOffsets(y, offset=offset)
    weights = _compressWeights(y, weights)
    for i in range(grid_npts):
        spline_disp = dispersion * 2 ** spline_pts[i]
        apl[:, i] = adjustedProfileLik(
            spline_disp, y=y, design=design, offset=offset, weights=weights
        )

    if trend:
        o = np.argsort(AveLogCPM)
        oo = np.argsort(o)
        width = floor(span * ntags)
        apl_smooth = movingAverageByCol(apl[o, :], width=width)[oo, :]
    else:
        apl_smooth = np.full((ntags, grid_npts), apl.mean(axis=0), order="F")
    apl_smooth = (apl + prior_n * apl_smooth) / (1 + prior_n)

    # Tagwise maximization
    d = maximizeInterpolant(spline_pts, apl_smooth)
    d = np.asarray(d, order="F")
    return dispersion * 2**d



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

cimport cython
from scipy.special cimport cython_special as sp

@cython.ufunc
cdef double nbinom_logpmf(double x, double n, double p):
    coeff = sp.gammaln(n+x) - sp.gammaln(x+1) - sp.gammaln(n)
    return coeff + sp.xlogy(n, p) + sp.xlog1py(x, -p)




# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/predFC.R' of the Bioconductor edgeR package (version 3.38.4).


import logging
import numpy as np

from .addPriorCount import addPriorCount


def predFC_DGEList(
    self, design, prior_count=0.125, offset=None, dispersion=None, weights=None
):
    """
    Compute estimated coefficients for a negative binomial GLM in such a way
    that the log-fold-changes are shrunk towards zero.

    See also
    --------
    predFC

    Arguments
    ---------
    """
    if offset is None:
        offset = self.getOffset()
    if dispersion is None:
        dispersion = self.getDispersion()
    if dispersion is None:
        dispersion = 0
        logging.warnings.warn("dispersion set to zero")

    return predFC(
        y=self.counts,
        design=design,
        prior_count=prior_count,
        offset=offset,
        dispersion=dispersion,
        weights=weights,
    )


def predFC(y, design, prior_count=0.125, offset=None, dispersion=0, weights=None):
    """
    Compute estimated coefficients for a negative binomial GLM in such a way
    that the log-fold-changes are shrunk towards zero.

    This function computes predictive log-fold changes (pfc) for a NB GLM. The
    pfc are posterior Bayesian estimators of the true log-fold-changes. They are
    predictive of values that might be replicated in a future experiment.

    Specifically, the function adds a small prior count to each observation
    before fitting the GLM (see :func:`addPriorCount` for details). The actual
    prior count that is added is proportional to the library size. This has the
    effect that any log-fold-change that was zero prior to augmentation remains
    zero and non-zero log-fold-changes are shrunk towards zero.

    The prior counts can be viewed as equivalent to a prior belief that the
    log-fold-changes are small, and the output can be viewed as posterior
    log-fold-changes from this Bayesian viewpoint. The output coefficients are
    called *predictive* log-fold-changes because, depending on the prior, They
    may be a better prediction of the true log-fold-changes than the raw
    estimates.

    Log-fold-changes for genes with low counts are shrunk more than those for
    genes with high counts. In particular, infinite log-fold-changes arising
    from zero counts are avoided. The exact degree to which this is done depends
    on the negative binomial dispersion.

    See also
    --------
    glmFit, exactTest, addPriorCount

    Arguments
    ---------
    y : array_like
        matrix of counts
    design : array_like
        the design matrix for the experiment
    prior_count : float
        the average prior count to be added to each observation. Larger values
        produce more shrinkage.
    offset : array_like
        vector or matrix giving the offset in the log-linear model predicto,
        as in :func:`glmFit`. Usually equal to log library size.
    dispersion : array_like
        vector of negative binomial dispersions
    weights : array_like, optional
        observation weights

    Returns
    -------
    ndarray
        matrix of (shrunk) linear model coefficients on the log2 scale

    References
    ----------
    B. Phipson. 2013. Empirical Bayes modelling of expression profiles and their
    associations. PhD thesis. University of Melbourne, Australia.
    http://repository.unimelb.edu.au/10187/17614
    """
    from .glmFit import glmFit

    # Add prior counts in proportion to library size
    (out_y, out_offset) = addPriorCount(y, offset=offset, prior_count=prior_count)

    # Check design
    design = np.asarray(design, order="F")

    # Return matrix of coefficients on log2 scale
    g = glmFit(
        out_y,
        design,
        offset=out_offset,
        dispersion=dispersion,
        prior_count=0,
        weights=weights,
    )
    return g.coefficients / np.log(2)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/nbinomDeviance.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from .makeCompressedMatrix import _compressDispersions, _compressWeights
from .edgepy_cpp import cxx_compute_nbdev_sum, cxx_compute_nbdev_nosum


def nbinomDeviance(y, mean, dispersion=0, weights=None):
    """
    Residual deviances for row-wise negative binomial GLMs.

    This function computes the total residual deviance for each row of :code:`y`,
    i.e. weighted row sums of the unit deviances.

    Care is taken to ensure accurate computation in limiting cases when the
    dispersion is near zero of :code:`mean * dispersion` is very large.

    See also
    --------
    nbinomUnitDeviance

    Arguments
    ---------
    y : array_like
        matrix containing the negative binomial counts, with rows for genes and
        columns for libraries. A vector will be treated as a matrix with one row.
    mean : array_like
        matrix of expected values, of same shape as :code:`y`. A vector will be
        treated as a matrix with one row.
    dispersion : array_like
        vector or matrix of negative binomial dispersions, as in :code:`glmFit`.
        Can be a scalar, a vector of length equal to the number of rows in
        :code:`y`, or a matrix of same shape as :code:`y`.
    weights : array_like, optional
        vector or matrix of non-negative weights, as in :code:`glmFit`. Can be
        a scalar, a vector of length equal to the number of columns in :code:`y`,
        or a matrix of same shape as :code:`y`.

    Returns
    -------
    ndarray
        vector of length equal to the number of rows in :code:`y`
    """
    out = _compute_nbdeviance(
        y=y, mean=mean, dispersion=dispersion, weights=weights, dosum=True
    )
    return out


def _compute_nbdeviance(y, mean, dispersion, weights, dosum):
    y = np.asarray(y, order="F")
    mean = np.asanyarray(mean, order="F", dtype="double")
    dispersion = np.asanyarray(dispersion, order="F", dtype="double")

    # Check y. May be matrix or vector
    if len(y.shape) == 2:
        if len(mean.shape) != 2:
            raise ValueError("y is a matrix but mean is not")
    else:
        n = y.shape[0]
        y = y.reshape((1, n))
        if len(mean.shape) == 2:
            raise ValueError("mean is a matrix but y is not")
        else:
            if mean.shape[0] == n:
                mean = mean.reshape((1, n))
            else:
                raise ValueError("length of mean differs from that of y")

        if len(dispersion.shape) == 2:
            raise ValueError("dispersion is a matrix but y is not")
        else:
            if len(dispersion.shape) > 0 and dispersion.shape[0] == n:
                dispersion = dispersion.reshape((1, n))
            else:
                raise ValueError("length of dispersion differs from that of y")

    # Check mean
    if y.shape != mean.shape:
        raise ValueError("mean should have the same dimensions as y")
    mean = mean.astype(np.float64, copy=False)

    # Check dispersion (can be tagwise (i.e. rowwise) or observation-wise)
    dispersion = _compressDispersions(y, dispersion)

    # Check weights
    weights = _compressWeights(y, weights)

    # Compute matrix of unit deviances, or residual deviance per gene, depending on `dosum`
    if dosum:
        return cxx_compute_nbdev_sum(y, mean, dispersion, weights)
    else:
        return cxx_compute_nbdev_nosum(y, mean, dispersion, weights)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/systematicSubset.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from math import floor


def systematicSubset(n, order_by):
    """
    Take a systematic subset of indices stratified by a ranking variable

    Arguments
    ---------
    n : int
        the size of the subset
    order_by : array_like
        vector of the values by which the indices are ordered

    Returns
    -------
    ndarray
        a vector of size :code:`n`
    """
    ntotal = len(order_by)
    sampling_ratio = floor(ntotal / n)
    if sampling_ratio <= 1:
        return np.arange(ntotal)
    i1 = floor(sampling_ratio / 2)
    i = np.arange(i1, ntotal, step=sampling_ratio)
    o = np.argsort(order_by)
    return o[i]



import unittest

from inmoose.utils import Factor
from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq, nbinomLRT


class Test(unittest.TestCase):
    def test_LRT(self):
        """test that test='LRT' gives correct errors"""
        dds = makeExampleDESeqDataSet(n=100, m=4)
        dds.obs["group"] = Factor([1, 2, 1, 2])
        dds.design = "~condition"
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="the following variables in the reduced formula not in the full formula: group",
        ):
            DESeq(dds, test="LRT", reduced="~group")
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="test='LRT' does not support use of expanded model matrix",
        ):
            DESeq(dds, test="LRT", reduced="~1", modelMatrixType="expanded")
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="test='LRT' does not support use of LFC shrinkage, use betaPrior=False",
        ):
            DESeq(dds, test="LRT", reduced="~group", betaPrior=True)
        dds = dds.estimateSizeFactors()
        dds = dds.estimateDispersions()
        with self.assertRaisesRegex(
            ValueError, expected_regex="provide a reduced formula for the LRT"
        ):
            nbinomLRT(dds)

    @unittest.skip("unimplemented")
    def test_glmGamPoi(self):
        raise NotImplementedError()

    def test_LRT2(self):
        """test that test='LRT' with quasi-likelihood estimates gives correct errors"""
        dds = makeExampleDESeqDataSet(n=100, m=4)
        dds.obs["group"] = Factor([1, 2, 1, 2])
        dds.design = "~condition + group"
        with self.assertRaises(NotImplementedError):
            DESeq(dds, test="Wald", fitType="glmGamPoi")
        dds = dds.estimateSizeFactors()
        dds_gp = dds.estimateDispersions()
        with self.assertRaises(NotImplementedError):
            nbinomLRT(dds_gp, reduced="~condition", type_="glmGamPoi")



import unittest
import numpy as np

from inmoose.edgepy.makeCompressedMatrix import makeCompressedMatrix


class Test(unittest.TestCase):
    def test_makeCompressedMatrix(self):
        self.assertTrue(
            np.array_equal(
                makeCompressedMatrix(42, dims=(2, 3), byrow=True), np.full((2, 3), 42)
            )
        )
        self.assertTrue(
            np.array_equal(
                makeCompressedMatrix(42, dims=(2, 3), byrow=False), np.full((2, 3), 42)
            )
        )

        ref = np.array([[1, 2, 3] for i in range(2)])
        self.assertTrue(
            np.array_equal(
                makeCompressedMatrix([1, 2, 3], dims=(2, 3), byrow=True), ref
            )
        )
        self.assertTrue(
            np.array_equal(
                makeCompressedMatrix([1, 2, 3], dims=(3, 2), byrow=False), ref.T
            )
        )

        with self.assertRaisesRegex(
            ValueError, expected_regex="dims\[.\] should be equal to length of x"
        ):
            makeCompressedMatrix([1, 2, 3], dims=(3, 2), byrow=True)
        with self.assertRaisesRegex(
            ValueError, expected_regex="dims\[.\] should be equal to length of x"
        ):
            makeCompressedMatrix([1, 2, 3], dims=(2, 3), byrow=False)
        with self.assertRaisesRegex(
            ValueError, expected_regex="dims does not represent the shape of a matrix"
        ):
            makeCompressedMatrix(42, dims=(5,))
        with self.assertRaisesRegex(
            ValueError, expected_regex="dims does not represent the shape of a matrix"
        ):
            makeCompressedMatrix(42, dims=(1, 2, 3))
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="input has too many dimensions to be interpreted as a matrix",
        ):
            makeCompressedMatrix(np.ones(shape=(1, 2, 3)), dims=(2, 3))


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import pandas as pd

from inmoose.utils import Factor
from inmoose.deseq2 import DESeqDataSet


class Test(unittest.TestCase):
    def test_method_errors(self):
        """test that methods throw errors"""
        coldata = pd.DataFrame({"x": Factor(["A", "A", "B", "B"])})
        counts = np.arange(1, 17).reshape((4, 4))
        dds = DESeqDataSet(counts, coldata, "~x")
        with self.assertWarnsRegex(
            UserWarning,
            expected_regex="there is no layer named 'replacedCounts', using original.",
        ):
            dds.counts(replaced=True)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="first calculate size factors, add normalizationFactors, or set normalized=False",
        ):
            dds.counts(normalized=True)
        with self.assertRaisesRegex(
            ValueError, expected_regex="size factors should be positive"
        ):
            dds.sizeFactors = [-1, -1, -1, -1]
        with self.assertRaisesRegex(
            ValueError, expected_regex="normalization factors should be positive"
        ):
            dds.normalizationFactors = np.full((4, 4), -1)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="first call estimateSizeFactors or provide a normalizationFactor matrix before calling estimateDispersions",
        ):
            dds.estimateDispersions()



import unittest
import numpy as np
import pandas as pd
import patsy
import scipy.stats

from inmoose.utils import Factor
from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq, DESeqDataSet


class Test(unittest.TestCase):
    def test_model_matrix(self):
        """test that supplying custom model matrix works"""
        dds = makeExampleDESeqDataSet(n=100, m=18)
        dds.obs["group"] = Factor(np.repeat([1, 2, 3], [6, 6, 6]))
        dds.obs["condition"] = Factor(
            np.repeat([np.repeat(["A", "B", "C"], [2, 2, 2])], 3, axis=0).flatten()
        )
        # note: design is not used
        dds.design = "~1"
        dds = dds[:16]

        m1 = patsy.dmatrix("~ group*condition", dds.obs)
        m2 = m1[:, :8]
        m2.design_info = m1.design_info
        m1 = m2
        m1.design_info.column_name_indexes.popitem()
        m0 = patsy.dmatrix("~ group + condition", dds.obs)

        # TODO
        # dds = DESeq(dds, full=m1, reduced=m0, test="LRT")
        # dds.result()
        # dds.results(name="group2.conditionC", test="Wald")
        dds = dds.removeResults()
        dds = DESeq(dds, full=m1, test="Wald", betaPrior=False)
        dds.results()

        # test better error than error during matrix inversion
        clindata = pd.DataFrame(
            {
                "group": Factor(np.repeat([1, 2, 3], 6)),
                "group2": Factor(np.repeat([1, 2, 3], 6)),
                "condition": Factor(np.repeat([1, 2, 3, 4, 5, 6], 3)),
            }
        )
        counts = scipy.stats.poisson.rvs(100, size=(18, 10))
        m1 = patsy.dmatrix("~ group + group2", clindata)
        m2 = patsy.dmatrix("~ condition + group", clindata)
        dds = DESeqDataSet(counts, clindata, "~ group")
        with self.assertRaisesRegex(
            ValueError, expected_regex="the model matrix is not full rank"
        ):
            dds = DESeq(dds, full=m1, fitType="mean")
        with self.assertRaisesRegex(
            ValueError, expected_regex="the model matrix is not full rank"
        ):
            dds = DESeq(dds, full=m2, reduced=m1, test="LRT", fitType="mean")



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/addPriorCount.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from .makeCompressedMatrix import _compressOffsets, _compressPrior, makeCompressedMatrix
from .edgepy_cpp import cxx_add_prior_count


def addPriorCount(y, lib_size=None, offset=None, prior_count=1):
    """
    Add a library size-adjusted prior count to each observation.

    This function adds a positive prior count to each observation, often useful
    to avoid zeroes during calculation of log-values. For example,
    :func:`predFC` will call this function to calculate shrunken log-fold
    changes. :func:`aveLogCPM` and :func:`cpm` also use the same underlying
    code to calculate (average) log-counts per million.

    The actual value added to the counts for each library is scaled according
    to the library size. This ensures that the relatives contribution of the
    prior is the same for each library. Otherwise, a fixed prior would have
    little effect on a large library, but a big effect on a small library.

    The library sizes are also modified, with twice the scales prior being
    added to the library size for each library. To understand the motication
    for this, consider that each observation is, effectively, a proportion of
    the total count in the library. The addition scheme implemented here
    represents an empirical logistic transform and ensures that the proportion
    can never be zero or one.

    If :code:`offset` is supplied, this is used in favor of :code:`lib_size`,
    where :code:`exp(offset)` is defined as the vector/matrix of library sizes.
    If an offset matrix is supplied, this will lead to gene-specific scaling of
    the prior as described above.

    Most use cases of this function will involve supplying a constant value to
    :code:`prior_count` for all genes. However, it is also possible to use
    gene-specific values by supplying a vector of length equal to the number of
    rows in :code:`y`.

    Arguments
    ---------
    y : matrix
        a numeric count matrix, with rows corresponding to genes and columns to
        libraries
    lib_size : array, optional
        a numeric vector of library sizes
    offset : array, optional
        a numeric vector or matrix of offsets
    prior_count : float or array
        a constant or gene-specific vector of prior counts to be added genes

    Returns
    -------
    matrix
        matrix of counts with the added priors
    CompressedMatrix
        the log-transformed modified library sizes
    """

    # Check y
    y = np.asarray(y, order="F")
    if not np.issubdtype(y.dtype, np.number):
        raise ValueError("count matrix must be numeric")

    # Check prior_count
    prior_count = _compressPrior(y, prior_count)

    # Check lib_size and offset
    # If offsets are provided, they must have a similar average to log(lib_size)
    # for the results to be meaningful as logCPM values
    offset = _compressOffsets(y, lib_size=lib_size, offset=offset)

    # Adding the prior count
    (out_y, out_offset) = cxx_add_prior_count(y, offset, prior_count)
    out_offset = makeCompressedMatrix(out_offset, y.shape, byrow=True)
    return (out_y, out_offset)



# -----------------------------------------------------------------------------
# Copyright (C) 2023 M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""
This module contains utilities to simulate RNASeq and single-cell RNASeq count
data. It follows the Splat model, as described in https://genomebiology.biomedcentral.com/articles/10.1186/s13059-017-1305-0.
Yet it is a completely independent implementation.
"""

import numpy as np
from scipy.stats import gamma, lognorm, chi2, poisson, bernoulli
from scipy.special import expit


def get_lognorm_factors(size, sel_prob, neg_prob, loc, scale, random_state):
    """
    Draw log-normal factors.

    Arguments
    ---------
    size : int or tuple of ints, optional
        Defining number of random variates
    sel_prob : float
        Probability that a factor is selected to be different of 1
    neg_prob : float
        Probability that a factor is less than 1
    loc : array_like, optional
        Location parameter of the log-normal distribution
    scale : array_like, optional
        Scale parameter of the log-normal distribution
    random_state : {int, `numpy.random.Generator` or `numpy.random.RandomState`},
                    optional
        If `random_state` is None (or `np.random`), the
        `numpy.random.RandomState` singleton is used.  If `random_state` is an
        int, a new ``RandomState`` instance is used, seeded with
        `random_state`.  If `random_state` is already a ``Generator`` or
        ``RandomState`` instance, that instance is used.

    Returns
    -------
    ndarray
        Random factors of given `size`
    """
    # make sure we have a Generator
    random_state = np.random.default_rng(random_state)

    selected = bernoulli.rvs(sel_prob, size=size, random_state=random_state) == 1
    n_selected = selected.sum()
    dir_selected = (-1) ** bernoulli.rvs(
        neg_prob, size=n_selected, random_state=random_state
    )
    # /!\ original R code uses `rlnorm(size, loc, scale)`, but the loc and scale correspond to the log-mean and log-sd
    #     therefore, the correct counterpart with scipy is `lognorm.rvs(s=scale, scale=np.exp(loc), size=size)`
    fac_selected = lognorm.rvs(
        s=scale, scale=np.exp(loc), size=n_selected, random_state=random_state
    )
    # reverse directions for factors that are less than one
    dir_selected[fac_selected < 1] *= -1
    factors = np.ones(size)
    factors[selected] = fac_selected**dir_selected
    return factors


def sim_rnaseq(
    nb_genes,
    nb_samples,
    batch=None,
    group=None,
    single_cell=False,
    alpha=0.6,
    beta=0.3,
    outlier_prob=0.05,
    outlier_location=4,
    outlier_scale=0.5,
    libsize_loc=11,
    libsize_scale=0.2,
    phi=0.1,
    bcv_df=60,
    x0=0,
    k=-1,
    random_state=None,
):
    """
    Simulate (sc)RNASeq data.

    For a precise description and understanding of the parameters, please refer
    to https://genomebiology.biomedcentral.com/articles/10.1186/s13059-017-1305-0.

    Arguments
    ---------
    nb_genes : int
        number of genes
    nb_samples : int
        number of samples
    batch : array-like, optional
        batch indices. Must have as many elements as `nb_samples`.
    group : array-like, optional
        vector/factor for biological condition of interest. Must have as many
        elements as `nb_samples`.
    dropout : bool, optional
        if True, simulate scRNASeq data by adding drop-outs. Defaults to `False`.

    alpha : float, optional
        shape of the gamma distribution to draw the initial means from
    beta : float, optional
        rate of the gamme distribution to draw the initial means from
    outlier_prob : float, optional
        proportion of outliers
    outlier_location : float, optional
        location parameter for the log-normal distribution to draw the outlier
        inflation factors from
    outlier_scale : float, optional
        scale parameter for the log-normal distribution to draw the outlier
        inflation factors from
    libsize_loc : float, optional
        mean of the normal distribution to draw the log of the library sizes
        from
    libsize_scale : float, optional
        standard deviation of the normal distribution to draw the log of the
        library sizes from
    phi : float, optional
        common dispersion
    bcv_df : int, optional
        degrees of freedom in the inverse chi-square distribution used to draw
        the biological coefficient of variation parameters
    x0 : float, optional
        dropout midpoint, used to draw the dropouts. Unused if
        `single_cell=False`.
    k : float, optional
        dropout shape, used to draw the dropouts. Unused if `single_cell=False`.
    random_state : {int, `numpy.random.Generator` or `numpy.random.RandomState`},
                    optional
        If `random_state` is None (or `np.random`), the
        `numpy.random.RandomState` singleton is used.  If `random_state` is an
        int, a new ``RandomState`` instance is used, seeded with
        `random_state`.  If `random_state` is already a ``Generator`` or
        ``RandomState`` instance, that instance is used.

    Returns
    -------
    ndarray
        simulated count matrix, of size `nb_genes` x `nb_samples`
    """

    # make sure we have a Generator
    random_state = np.random.default_rng(random_state)

    if batch is None:
        batch = np.zeros(nb_samples, dtype=np.int64)
    batch = np.asarray(batch)
    assert len(batch) == nb_samples

    if group is None:
        group = np.zeros(nb_samples, dtype=np.int64)
    group = np.asarray(group)
    assert len(group) == nb_samples

    nb_batches = len(np.unique(batch))
    nb_groups = len(np.unique(group))

    # draw gene means
    original_means = gamma.rvs(
        alpha, scale=1 / beta, size=nb_genes, random_state=random_state
    )

    # draw outliers
    outlier_factor = get_lognorm_factors(
        original_means.shape,
        outlier_prob,
        0,
        outlier_location,
        outlier_scale,
        random_state,
    )
    is_outlier = outlier_factor != 1.0
    gene_means = np.where(
        is_outlier, outlier_factor * np.median(original_means), original_means
    )

    # draw library sizes
    L = lognorm.rvs(
        libsize_scale,
        scale=np.exp(libsize_loc),
        size=nb_samples,
        random_state=random_state,
    )

    # cell means
    cell_means = L * gene_means[:, None] / gene_means.sum()

    # biological coefficients of variation (squared)
    B2 = (phi + 1 / np.sqrt(cell_means)) ** 2 * (
        bcv_df / chi2.rvs(bcv_df, size=nb_genes, random_state=random_state)
    )[:, None]

    # trended cell means
    trended_cell_means = gamma.rvs(1 / B2, cell_means * B2, random_state=random_state)

    # DE fold-changes between cell types
    de_lf = [
        get_lognorm_factors(
            nb_genes, 1, 0.3 + 0.2 * i, 0.1 + i * 0.5, 1, random_state=random_state
        )
        for i in range(nb_groups)
    ]
    de_lf = np.choose(group, [x[:, None] for x in de_lf])

    # cell type specific cell means
    de_trended_cell_means = de_lf * trended_cell_means

    # batch effect factors
    be_fact = [
        get_lognorm_factors(
            nb_genes, 1, 0.5, 0.1 + 0.5 * i, 1, random_state=random_state
        )
        for i in range(nb_batches)
    ]
    be_fact = np.choose(batch, [x[:, None] for x in be_fact])
    batch_cell_means = be_fact * de_trended_cell_means

    # counts
    y = poisson.rvs(batch_cell_means, random_state=random_state)

    if single_cell:
        # simulate drop-outs
        pi = expit(k * (np.log(batch_cell_means) - x0))
        y = bernoulli.rvs(pi, random_state=random_state) * y

    return y



import unittest
import numpy as np
import patsy

from inmoose.utils import Factor, pt
from inmoose.deseq2 import (
    makeExampleDESeqDataSet,
    nbinomWaldTest,
    nbinomLRT,
    DESeq,
    estimateMLEForBetaPriorVar,
    estimateBetaPriorVar,
    estimateDispersionsGeneEst,
)


class Test(unittest.TestCase):
    def test_nbinomWald_errors(self):
        """test that nbinomWald throws various errors and works with edge cases"""
        dds = makeExampleDESeqDataSet(n=100, m=4)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="testing requires dispersion estimates, first call estimateDispersions()",
        ):
            nbinomWaldTest(dds)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="testing requires dispersion estimates, first call estimateDispersions()",
        ):
            nbinomLRT(dds)

        dds = dds.estimateSizeFactors()
        dds = dds.estimateDispersions()
        mm = patsy.dmatrix("~condition", dds.obs)
        mm0 = patsy.dmatrix("~1", dds.obs)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="user-supplied model matrix with betaPrior=True requires supplying betaPriorVar",
        ):
            nbinomWaldTest(dds, betaPrior=True, modelMatrix=mm)
        # TODO
        # with self.assertRaisesRegex(ValueError, expected_regex="unused argument (betaPrior = TRUE)"):
        #    nbinomLRT(dds, betaPrior=True, full=mm, reduced=mm0)
        with self.assertRaisesRegex(
            ValueError, expected_regex="expanded model matrices require a beta prior"
        ):
            nbinomWaldTest(dds, betaPrior=False, modelMatrixType="expanded")
        # with self.assertRaisesRegex(ValueError, expected_regex="unused arguments (betaPrior = FALSE, modelMatrixType = 'expanded')"):
        #    nbinomLRT(dds, betaPrior=False, modelMatrixType="expanded")

        dds2 = estimateMLEForBetaPriorVar(dds.copy())
        estimateBetaPriorVar(dds2, betaPriorMethod="quantile")
        dds = nbinomWaldTest(dds, modelMatrixType="standard")
        # TODO
        # covarianceMatrix(dds, 1)

        # changing 'df'
        dds = makeExampleDESeqDataSet(n=100, m=4)
        dds.X[:, :4] = 0
        dds = dds.estimateSizeFactors()
        dds = dds.estimateDispersions()
        dds = nbinomWaldTest(dds)
        dds.results().pvalue[:8]
        dds = nbinomWaldTest(dds, useT=True, df=np.ones(100))
        dds.results().pvalue[:8]

        # try nbinom after no fitted dispersions
        dds = makeExampleDESeqDataSet(n=100, m=4)
        dds = dds.estimateSizeFactors()
        dds = estimateDispersionsGeneEst(dds)
        dds.var["dispersion"] = dds.var["dispGeneEst"]
        dds = nbinomWaldTest(dds)

    def test_nbinomWald_useT(self):
        """test that useT uses proper degrees of freedom"""
        dds = makeExampleDESeqDataSet(n=200, m=15, seed=42)
        dds.X[:, 100:105] = 0
        dds.obs["condition"] = Factor(np.repeat(["A", "B", "C"], 5))
        dds.design = "~condition"
        dds = DESeq(dds, useT=True)
        dds = dds.removeResults()
        w = np.ones(dds.shape)
        w[0, :100] = 0
        w[:4, 0] = 0
        w[5:9, 0] = 0
        w[10:14, 0] = 0
        dds.layers["weights"] = w
        dds = DESeq(dds, useT=True)
        res = dds.results()
        self.assertTrue(np.isnan(res.pvalue[0]))
        self.assertEqual(dds.var["tDegreesFreedom"][1], 15 - 1 - 3)
        self.assertEqual(
            res.pvalue[1], 2 * pt(np.abs(res.stat[1]), df=15 - 1 - 3, lower_tail=False)
        )

        # also lfcThreshold
        res = dds.results(lfcThreshold=1, altHypothesis="greaterAbs")
        idx = np.nonzero(((res.log2FoldChange > 1) & ~np.isnan(res.pvalue)).values)[0][
            0
        ]
        self.assertEqual(
            res.pvalue[idx], 2 * pt(res.stat[idx], df=15 - 1 - 3, lower_tail=False)
        )
        res = dds.results(lfcThreshold=1, altHypothesis="greater")
        idx = np.nonzero(((res.log2FoldChange > 1) & ~np.isnan(res.pvalue)).values)[0][
            0
        ]
        self.assertEqual(
            res.pvalue[idx], pt(res.stat[idx], df=15 - 1 - 3, lower_tail=False)
        )

        res = dds.results(lfcThreshold=1, altHypothesis="less")
        idx = np.nonzero(((res.log2FoldChange < -1) & ~np.isnan(res.pvalue)).values)[0][
            0
        ]
        self.assertEqual(
            res.pvalue[idx], pt(-res.stat[idx], df=15 - 1 - 3, lower_tail=False)
        )

        res = dds.results(lfcThreshold=1, altHypothesis="lessAbs")
        idx = np.nonzero(
            ((np.abs(res.log2FoldChange) < 1) & ~np.isnan(res.pvalue)).values
        )[0][0]
        self.assertEqual(
            res.pvalue[idx], pt(res.stat[idx], df=15 - 1 - 3, lower_tail=False)
        )

        # also novel contrasts
        res = dds.results(contrast=["condition", "C", "B"])
        self.assertTrue(np.isnan(res.pvalue[0]))
        self.assertTrue(dds.var["tDegreesFreedom"][1] == 15 - 1 - 3)
        self.assertTrue(
            res.pvalue[1] == 2 * pt(abs(res.stat[1]), df=15 - 1 - 3, lower_tail=False)
        )



import unittest
from inmoose.consensus_clustering.consensus_clustering import consensusClustering
from sklearn.cluster import AgglomerativeClustering
import pandas as pd
import numpy as np
import importlib.resources
import os


class test_consensusClustering(unittest.TestCase):
    def setUp(self) -> None:
        this_dir = importlib.resources.files(__package__)
        mock_file = this_dir.joinpath("mocked_data_consensus_clustering.csv")

        # Fix random seeds to always get the same results in the tests
        self.CC = consensusClustering(
            cluster=AgglomerativeClustering,
            mink=2,
            maxk=4,
            nb_resampling_iteration=50,
            resample_proportion=0.5,
        )
        self.mocked_data = pd.read_csv(mock_file, index_col=0)

        self.deltak_plot = this_dir.joinpath("deltak_plot.png")
        self.clustermap_plot = this_dir.joinpath("clustermap_plot.png")
        self.consensus_plot = this_dir.joinpath("clusters_consensus_plot.png")
        self.consensus_line_plot = this_dir.joinpath("clusters_consensus_line_plot.png")

    def tearDown(self) -> None:
        if os.path.exists(self.deltak_plot):
            os.remove(self.deltak_plot)

        if os.path.exists(self.clustermap_plot):
            os.remove(self.clustermap_plot)

        if os.path.exists(self.consensus_plot):
            os.remove(self.consensus_plot)

        if os.path.exists(self.consensus_line_plot):
            os.remove(self.consensus_line_plot)

    def test_internal_resample(self):
        resampled_indices = self.CC._internal_resample(
            self.mocked_data, 0.5, np.random.default_rng()
        )

        assert len(resampled_indices) == len(self.mocked_data) / 2

    def test_compute_consensus_clustering(self):
        self.CC.compute_consensus_clustering(
            self.mocked_data.to_numpy(), random_state=0
        )

        # test consensus matrix is symetric
        for i in range(len(self.CC.consensus_matrices)):
            assert np.allclose(
                self.CC.consensus_matrices[i], self.CC.consensus_matrices[i].T
            )

        # test max consensus matrices = 1 and min = 0
        assert np.max(self.CC.consensus_matrices) == 1
        assert np.min(self.CC.consensus_matrices) == 0

        # assert bestK
        assert self.CC.bestK == 3

    def test_plot_clustermap(self):
        self.CC.compute_consensus_clustering(
            self.mocked_data.to_numpy(), random_state=0
        )
        self.CC.plot_clustermap(3, self.clustermap_plot)
        assert os.path.exists(self.clustermap_plot)

    def test_plot_deltak(self):
        self.CC.compute_consensus_clustering(
            self.mocked_data.to_numpy(), random_state=0
        )
        self.CC.plot_deltak(self.deltak_plot)
        assert os.path.exists(self.deltak_plot)

    def test_plot_clusters_consensus(self):
        self.CC.compute_consensus_clustering(
            self.mocked_data.to_numpy(), random_state=0
        )
        cons_clust_df = self.CC.build_clusters_consensus_df()
        self.CC.plot_clusters_consensus(cons_clust_df, self.consensus_plot)
        assert os.path.exists(self.consensus_plot)

    def test_line_plots_cluster_consensus(self):
        self.CC.compute_consensus_clustering(
            self.mocked_data.to_numpy(), random_state=0
        )
        cons_clust_df = self.CC.build_clusters_consensus_df()
        self.CC.line_plots_cluster_consensus(cons_clust_df, self.consensus_line_plot)
        assert os.path.exists(self.consensus_line_plot)



def airway():
    """
    Retrieve the airway dataset as an :class:`AnnData` object

    The airway dataset is an RNA-Seq experiment on four human airway smooth
    muscle cell lines treated with dexamethasone [Himes2014]_.
    """
    import anndata
    import importlib.resources

    data_dir = importlib.resources.files(__package__)
    return anndata.read_h5ad(data_dir.joinpath("airway.h5ad"))



import unittest

from inmoose.data.pasilla import pasilla


class Test(unittest.TestCase):
    def test_pasilla(self):
        """
        check that the pasilla dataset can be properly loaded
        """
        adata = pasilla()

        self.assertEqual(adata.shape, (7, 14599))
        self.assertEqual(len(adata.obs.columns), 5)



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import logging
import numpy as np


def estimateSizeFactors_dds(
    obj,
    type_="ratio",
    locfunc=np.median,
    geoMeans=None,
    controlGenes=None,
    normMatrix=None,
    quiet=False,
):
    """
    Estimate the size factors of a :class:`DESeqDataSet`

    This function estimates the size factors using the "median ratio method",
    described by Equation 5 in [Anders2010]_.

    The estimated size factors can be accessed through the
    :attr:`DESeqDataSet.sizeFactors` property of :class:`DESeqDataSet`.
    Alternative library size estimators can also be supplied through this
    property.

    See :func:`.DESeq` for a description of the use of size factors in the GLM.
    One should call this function after building a :obj:`DESeqDataSet`, unless
    size factors are manually specified with property
    :attr:`DESeqDataSet.sizeFactors`.  Alternatively, gene-specific
    normalization factors for each sample can be provided using the
    :attr:`DESeqDataSet.normalizationFactors` which will always preempt
    :attr:`DESeqDataSet.sizeFactors` in calculations.

    Internally, the function calls :func:`.estimateSizeFactorsForMatrix`, which
    provides more details on the calculation.

    See also
    --------
    .DESeq
    .estimateSizeFactorsForMatrix

    Arguments
    ---------
    obj : DESeqDataSet
        the input dataset
    type_ : "ratio", "poscounts" or "iterate"
        the algorithm to estimate the size factors

        :code:`"ratio"` uses the standard median ratio method introduced in
        DESeq. The size factor is the median ratio of the sample over a
        "pseudosample": for each gene, the geometric mean of all samples.

        :code:`"poscounts"` and :code:`"iterate"` offer alternative estimators,
        which can be used even when all genes contain a sample with a zero (a
        problem for the default method, as the geometric then becomes zero, and
        the ratio undefined).

        The :code:`"poscounts"` estimator deals with a gene with some zeros by
        calculating a modified geometric mean by taking the n-th root of the
        product of the non-zero counts.  This evolved out of use cases with
        Paul McMurdie's phyloseq package for metagenomic samples.

        The :code:`"iterate"` estimator iterates between estimating the
        dispersion with a design of ~1, and finding a size factor vector by
        numerically optimizing the likelihood of the ~1 model.
    locfunc
        a function to compute a location for a sample. By default, the median is
        used.
    geoMeans : array-like, optional
        by default, the geometric means of the counts are calculated within the
        function. A vector of geometric means from another count matrix can be
        provided for a "frozen" size factor calculation. The size factors will
        be scaled to have a geometric mean of 1 when supplying :code:`geoMeans`.
    controlGenes : array-like, optional
        index vector specifying those genes to use for size factor estimation
        (e.g. housekeeping or spike-in genes)
    normMatrix : ndarray, optional
        a matrix of normalization factors which do not yet control for library
        size. Providing :code:`normMatrix` will estimate size factors on the
        count matrix divided by :code:`normMatrix` and store the product of the
        size factors and :code:`normMatrix` as
        :attr:`DESeqDataSet.normalizationFactors`.  It is recommended to divide
        out the sample-wise geometric mean of :code:`normMatrix` so the
        sample-wise factors are roughly centered on 1.
    quiet : bool
        controls verbosity, defaults to False

    Returns
    -------
    DESeqDataSet
        the input :code:`obj`, with the size factors filled in
    """

    if type_ not in ["ratio", "poscounts", "iterate"]:
        raise ValueError(f"invalid type: {type_}")

    if type_ == "iterate":
        raise NotImplementedError(
            "iterative method to estimate size factors not implemented yet"
        )
        obj.sizeFactors = obj.estimateSizeFactorsIterate()
    else:
        if type_ == "poscounts":

            def geoMeanNZ(x):
                if (x == 0).all():
                    return 0
                else:
                    return np.exp(np.sum(np.log(x[x > 0])) / len(x))

            geoMeans = np.apply_along_axis(geoMeanNZ, 0, obj.X)

        if "avgTxLength" in obj.obsm.keys():
            nm = obj.obsm["avgTxLength"]
            nm = nm / np.exp(np.mean(np.log(nm), 1))
            obj.normalizationFactors = estimateNormFactors(
                obj.X,
                normMatrix=nm,
                locfunc=locfunc,
                geoMeans=geoMeans,
                controlGenes=controlGenes,
            )
            logging.info(
                "using 'avgTxLength' from dds.obsm, correcting for library size"
            )

        elif normMatrix is None:
            obj.sizeFactors = estimateSizeFactorsForMatrix(
                obj.X, locfunc=locfunc, geoMeans=geoMeans, controlGenes=controlGenes
            )
        else:
            obj.normalizationFactors = estimateNormFactors(
                obj.X,
                normMatrix=normMatrix,
                locfunc=locfunc,
                geoMeans=geoMeans,
                controlGenes=controlGenes,
            )
            logging.info(
                "using 'normMatrix', adding normalization factors which correct for library size"
            )

    return obj


def estimateSizeFactorsForMatrix(
    counts, type_="ratio", locfunc=np.median, geoMeans=None, controlGenes=None
):
    """
    Low-level function to estimate size factors with robust regression

    Given a matrix or data frame of count data, this function estimates the
    size factors as follows: each row is divided by the geometric means of the
    columns. The median (or, if requested, another location estimator) of these
    ratios (skipping the genes with a geometric mean of zero) is used as the
    size factor for this row. Typically, one will not call this function
    directly, but use :meth:`.DESeqDataSet.estimateSizeFactors`.

    See also
    --------
    .DESeqDataSet.estimateSizeFactors

    Arguments
    ---------
    counts : array-like
        matrix of raw counts. One column per gene, one row per sample.
    type_ : "ratio" or "poscounts"
        the algorithm to estimate the size factors: standard median ratio
        (:code:`"ratio"`), or there the geometric mean is only calculated over
        positive counts (:code:`"poscounts"`).
    locfunc
        a function to compute a location for a sample. By default, the median
        is used.
    geoMeans : ndarray, optional
        by default, the geometric means of the counts are calculated within
        the function. A vector of geometric means from another count matrix can
        be provided for a "frozen" size factor calculation.
    controlGenes : array-like, optional
        index vector specifying those genes to use for size factor estimation
        (e.g. housekeeping or spike-in genes)

    Returns
    -------
    ndarray
        the estimated size factors, one element per row of :code:`counts`
    """

    if type_ not in ["ratio", "poscounts"]:
        raise ValueError(f"type_ must be either 'ratio' or 'poscounts': {type_}")

    if geoMeans is None:
        incomingGeoMeans = False
        if type_ == "ratio":
            loggeomeans = np.mean(np.log(counts), 0)
        elif type_ == "poscounts":
            lc = np.log(counts)
            lc[~np.isfinite(lc)] = 0
            loggeomeans = np.mean(lc, 0)
            allZero = np.sum(counts, 0) == 0
            loggeomeans[allZero] = -np.inf

    else:
        incomingGeoMeans = True
        if len(geoMeans) != counts.shape[1]:
            raise ValueError(
                "geoMeans should be as long as the number of columns of counts"
            )
        loggeomeans = np.log(geoMeans)

    if np.isinf(loggeomeans).all():
        raise ValueError(
            "every gene contains at least one zero, cannot compute log geometric means"
        )

    if controlGenes is None:

        def sf_compute(cnts):
            return np.exp(
                locfunc(
                    (np.log(cnts) - loggeomeans)[np.isfinite(loggeomeans) & (cnts > 0)]
                )
            )

        sf = np.apply_along_axis(sf_compute, 1, counts)
    else:
        loggeomeansSub = loggeomeans[controlGenes]

        def sf_compute(cnts):
            return np.exp(
                locfunc(
                    (np.log(cnts) - loggeomeansSub)[
                        np.isfinite(loggeomeansSub) & (cnts > 0)
                    ]
                )
            )

        sf = np.apply_along_axis(sf_compute, 1, counts[:, controlGenes])

    if incomingGeoMeans:
        # stabilize the size factors to have geometric mean of 1
        sf = sf / np.exp(np.mean(np.log(sf)))

    return sf


def estimateNormFactors(
    counts, normMatrix, locfunc=np.median, geoMeans=None, controlGenes=None
):
    sf = estimateSizeFactorsForMatrix(
        counts / normMatrix,
        locfunc=locfunc,
        geoMeans=geoMeans,
        controlGenes=controlGenes,
    )
    nf = normMatrix * sf.reshape(sf.shape[0], 1)
    return nf / np.exp(np.mean(np.log(nf), 0))



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import logging
import numpy as np


def getAndCheckWeights(obj, modelMatrix, weightThreshold=1e-2):
    """
    Check and retrive weights

    Weights checking consists in verifying that the model matrix remains full
    rank, even when considering weights below `weightThreshold` are considered
    to be zero.  Checking sets an attribute in `obj`, so that this verification
    is performed only once.

    Arguments
    ---------
    obj : DESeqDataSet
        the DESeqDataSet to retrieve weights from
    modelMatrix : matrix
        the design matrix
    weightThreshold : float
        the threshold below which weights may be considered zero

    Returns
    -------
    obj : DESeqDataSet
        the updated input `obj`
    weights : ndarray
        the matrix of weights (same shape as `obj`)
    useWeights : bool
        a flag indicating weights are used or not
    """
    if "weights" in obj.layers:
        useWeights = True
        weights = obj.layers["weights"]
        if not (weights >= 0).all():
            raise ValueError("weights must be positive")
        weights = weights / np.max(weights, 0)
        # some code for testing whether still full rank
        # only performed once per analysis, by setting obj attribute
        if obj.weightsOK is None:
            m = modelMatrix.shape[1]
            full_rank = np.linalg.matrix_rank(modelMatrix) == m
            weights_ok = np.repeat(False, weights.shape[1])
            # most designs are full rank with current version
            if full_rank:
                for i in range(weights.shape[1]):
                    # note: downweighting of samples very low will still be full rank
                    # so this test is king of minimally in play -- good for checking
                    # the user input however, e.g. all zero weights for a gene
                    test1 = (
                        np.linalg.matrix_rank(weights[:, i][:, None] * modelMatrix) == m
                    )
                    # we test that it will be possible to calculate the CR term
                    # following subsetting based on weightThreshold
                    mm_sub = modelMatrix[weights[:, i] > weightThreshold]
                    mm_sub = mm_sub[:, np.sum(np.abs(mm_sub), 0) > 0]
                    test2 = np.linalg.matrix_rank(mm_sub) == mm_sub.shape[1]
                    weights_ok[i] = test1 and test2
            else:
                # model matrix is not full rank (backwards compatibility for betaPrior=True)
                # just check for zero columns
                weights_ok = np.repeat(True, weights.shape[1])
                for j in range(modelMatrix.shape[1]):
                    num_zero = np.sum(weights.T * modelMatrix[:, j] == 0, axis=1)
                    weights_ok = weights_ok & (num_zero != modelMatrix.shape[0])

            # instead of giving an error, switch allZero to True for the problem cols
            if not weights_ok.all():
                obj.var["allZero"][~weights_ok] = True
                obj.var["weightsFail"] = ~weights_ok
                obj.var.type["weightsFail"] = "intermediate"
                obj.var.description[
                    "weightsFail"
                ] = "weights fail to allow parameter estimation"
                logging.warnings.warn(
                    f"for {np.sum(~weights_ok)} genes, the weights as supplied won't allow parameter estimation, producing a degenerate design matrix. These columns have been flagged in dds.var['weightsFail'] and treated as if the column contained all zeros (dds.var['allZero'] set to True). If you are blocking for donors/organisms, consider design = ~0+donor+condition."
                )

        obj.weightsOK = True

    else:
        useWeights = False
        weights = np.ones(obj.shape)

    return (obj, weights, useWeights)



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import numpy as np

from ..utils import Factor


def wtd_quantile(
    x, weights=None, probs=[0, 0.25, 0.5, 0.75, 1], normwt=False, na_rm=True
):
    """
    compute weighted quantiles

    This function is copied from R package `Hmisc
    <https://hbiostat.org/r/hmisc/>`_, to avoid extra package dependencies,
    with the alteration of leaving out parameter :code:`type` (not used in
    DESeq) and commenting out the :code:`isdate` test.

    Note that this copy was already done in the original DESeq2 R code.
    """
    if np.any((probs < 0) | (probs > 1)):
        raise ValueError("Probabilities must be between 0 and 1 inclusive")

    if weights is None:
        if na_rm:
            return np.nanquantile(x, q=probs)
        else:
            return np.quantile(x, q=probs)

    i = np.isnan(weights) | weights == 0
    if np.any(i):
        x = x[~i]
        weights = weights[~i]

    (x, wts) = wtd_table(x, weights, na_rm=na_rm, normwt=normwt)
    n = np.sum(wts)
    order = 1 + (n - 1) * probs
    low = np.maximum(np.floor(order), 1)
    high = np.minimum(low + 1, n)
    order = order % 1
    ## Find low and high order statistics
    ## These are the minimum values of x such that the cum. freqs >= (low,high)
    allq = np.piecewise(
        np.hstack([low, high]),
        True,
        [lambda v: np.hstack([x, [x[-1]]])[np.searchsorted(np.cumsum(wts), v)]],
    )
    try:
        k = len(probs)
    except TypeError:
        k = 1
    quantiles = (1 - order) * allq[:k] + order * allq[-k]
    return quantiles


def wtd_table(x, weights=None, normwt=False, na_rm=True):
    if weights is None:
        weights = np.ones(len(x))

    x = Factor(x)
    lev = x.categories
    x = x.codes

    if na_rm:
        s = ~np.isnan(x + weights)
        x = x[s]
        weights = weights[s]

    if normwt:
        weights = weights * len(x) / np.sum(weights)

    i = np.argsort(x)
    x = x[i]
    weights = weights[i]

    if len(np.unique(x)) != len(x):
        x = np.asarray(x)
        weights = np.array([np.sum(weights[x == xx]) for xx in np.unique(x)])
        if len(lev) > 0:
            levused = lev[np.sort(np.unique(x))]
            if len(weights) > len(levused) and np.any(np.isnan(weights)):
                weights = weights[~np.isnan(weights)]

            if len(weights) != len(levused):
                raise RuntimeError("program logic error")

        return (levused, weights)

    if len(lev) > 0:
        return (lev[x], weights)
    else:
        return (x, weights)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/dispCoxReid.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from scipy.optimize import minimize_scalar

from .adjustedProfileLik import adjustedProfileLik
from .aveLogCPM import aveLogCPM
from .makeCompressedMatrix import _compressOffsets, _compressWeights
from .systematicSubset import systematicSubset


def dispCoxReid(
    y,
    design=None,
    offset=None,
    weights=None,
    AveLogCPM=None,
    interval=(0, 4),
    tol=1e-5,
    min_row_sum=5,
    subset=10000,
):
    """
    Estimate a common dispersion parameter across multiple negative binomial
    GLMs, by maximizing the Cox-Reid adjusted profile likelihood.

    This is a low-level function called by :func:`estimateGLMCommonDisp`.

    Estimation is done by maximizing the Cox-Reid adjusted profile likelihood
    (Cox and Reid, 1987 [1]_), through :func:`scipy.optimize.minimize_scalar`.

    Robinson and Smyth (2008) [2]_ and McCarthy et al. (2012) [3]_ showed that
    the Pearson (pseudo-likelihood) estimator typically under-estimates the true
    dispersion. It can be seriously biased when the number of libraries is small.
    On the other hand, the deviance (quasi-likelihood) estimator typically
    over-estimates the true dispersion when the number of libraries is small.
    Robinson and Smyth (2008) [2]_ and McCarthy et al. (2012) [3]_ showed the
    Cox-Reid estimator to be the least biased of the three options.

    Arguments
    ---------
    y : matrix
        matrix of counts. A GLM is fitted to each row
    design : matrix, optional
        design matrix, as in :func:`glmFit`
    offset : array_like, optional
        vector or matrix of offsets for the log-linear models, as in
        :func:`glmFit`. Defaults to :code:`log(colSums(y))`.
    weights : matrix, optional
        observation weights
    AveLogCPM : array_like, optional
        vector giving average log2 counts per million
    interval : tuple, optional
        pair giving minimum and maximum allowed values for the dispersion,
        passed to :func:`scipy.optimize.minimize_scalar`
    tol : float, optional
        the desired accuracy, see :func:`scipy.optimize.minimize_scalar`
    min_row_sum : int, optional
        only rows with at least this number of counts are used
    subset : int, optional
        number of rows to use in the calculation. Rows used are chosen evenly
        space by :code:`AveLogCPM`.

    Returns
    -------
    float
        the estimated common dispersion

    References
    ----------
    .. [1] D. R. Cox, N. Reid. 1987. Parameter orthogonality and approximate
       conditional inference. Journal of the Royal Statistical Society Series B
       49, 1-39.
    .. [2] M. D. Robinson, G. K. Smyth. 2008. Small-sample estimation of
       negative binomial dispersion, with applications to SAGE data.
       Biostatistics 9, 321-332.
    .. [3] D. J. McCarthy, Y. Chen, G. K. Smyth. 2012. Differential expression
       analysis of multifactor RNA-Seq experiments with respect to biological
       variation. Nucleic Acids Research 40, 4288-4297. :doi:`10.1093/nar/gks042`
    """
    # Check y
    y = np.asarray(y, order="F")

    # Check design
    if design is None:
        design = np.ones((y.shape[1], 1), order="F")
    else:
        design = np.asarray(design, order="F")

    # Check offset
    if offset is None:
        offset = np.log(y.sum(axis=0))
    if len(offset.shape) == 1:
        offset = np.full(y.shape, offset, order="F")
    assert offset.shape == y.shape

    if interval[0] < 0:
        raise ValueError("please give a non-negative interval for the dispersion")

    if AveLogCPM is not None:
        AveLogCPM = np.asarray(AveLogCPM, order="F")

    # Apply min row count
    small_row_sum = y.sum(axis=1) < min_row_sum
    if small_row_sum.any():
        y = y[np.logical_not(small_row_sum)]
        offset = offset[np.logical_not(small_row_sum)]
        if weights is not None:
            weights = weights[np.logical_not(small_row_sum)]
        if AveLogCPM is not None:
            AveLogCPM = AveLogCPM[np.logical_not(small_row_sum)]
    if y.shape[0] < 1:
        raise ValueError("no data rows with required number of counts")

    # Subsetting
    if subset is not None and subset <= y.shape[0] / 2:
        if AveLogCPM is None:
            AveLogCPM = aveLogCPM(y, offset=offset, weights=weights)
        i = systematicSubset(subset, AveLogCPM)
        y = y[i, :]
        offset = offset[i, :]
        if weights is not None:
            weights = weights[i, :]

    # Function for optimizing
    def sumAPL(par, y, design, offset, weights):
        return -sum(adjustedProfileLik(par**4, y, design, offset, weights=weights))

    # anticipate the calls to _compress* in adjustedProfileLik
    y = np.asarray(y, order="F")
    offset = _compressOffsets(y, offset=offset)
    weights = _compressWeights(y, weights)
    out = minimize_scalar(
        sumAPL,
        args=(y, design, offset, weights),
        bounds=(interval[0] ** 0.25, interval[1] ** 0.25),
        options={"xatol": tol},
    )
    return out.x**4



import unittest
import numpy as np

from inmoose.edgepy import pnbinom, qnbinom


class test_nbinom(unittest.TestCase):
    def test_nbinom(self):
        for i in range(10):
            hi = pnbinom(i, size=5, prob=0.2, lower_tail=False)
            lo = pnbinom(i, size=5, prob=0.2, lower_tail=True)
            self.assertEqual(hi, 1.0 - lo)
            self.assertEqual(i, qnbinom(hi, size=5, prob=0.2, lower_tail=False))
            self.assertEqual(i, qnbinom(lo, size=5, prob=0.2, lower_tail=True))

            hi = pnbinom(i, size=5, mu=4, lower_tail=False)
            lo = pnbinom(i, size=5, mu=4, lower_tail=True)
            self.assertAlmostEqual(hi, 1.0 - lo)
            self.assertEqual(i, qnbinom(hi, size=5, mu=4, lower_tail=False))
            self.assertEqual(i, qnbinom(lo, size=5, mu=4, lower_tail=True))

        ctxt = self.assertRaisesRegex(
            ValueError, expected_regex="exactly one of prob and mu must be provided"
        )
        with ctxt:
            pnbinom(42, size=5, prob=0.5, mu=4)
        with ctxt:
            qnbinom(42, size=5, prob=0.5, mu=4)
        with ctxt:
            pnbinom(42, size=5)
        with ctxt:
            qnbinom(42, size=5)



import sys
import numpy
from setuptools import Extension, setup

cxx_compile_flags = ["-std=c++17"]
macros = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
profiling = False
linetrace = False
if "--profile" in sys.argv:
    profiling = True
    linetrace = True
    macros += [("CYTHON_TRACE_NOGIL", "1")]
    sys.argv.remove("--profile")

common_cpp = Extension(
    "inmoose.common_cpp",
    [
        "inmoose/common_cpp/common_cpp.pyx",
        "inmoose/common_cpp/matrix.cpp",
    ],
    include_dirs=[numpy.get_include()],
    extra_compile_args=cxx_compile_flags,
    define_macros=macros,
)

stats_cpp = Extension(
    "inmoose.utils.stats_cpp",
    [
        "inmoose/utils/_stats.pyx",
    ],
    include_dirs=[numpy.get_include()],
    extra_compile_args=cxx_compile_flags,
    define_macros=macros,
)

edgepy_cpp = Extension(
    "inmoose.edgepy.edgepy_cpp",
    [
        "inmoose/edgepy/edgepy_cpp/edgepy_cpp.pyx",
        "inmoose/edgepy/edgepy_cpp/add_prior.cpp",
        "inmoose/edgepy/edgepy_cpp/add_prior_count.cpp",
        "inmoose/edgepy/edgepy_cpp/adj_coxreid.cpp",
        "inmoose/edgepy/edgepy_cpp/fit_levenberg.cpp",
        "inmoose/edgepy/edgepy_cpp/glm_levenberg.cpp",
        "inmoose/edgepy/edgepy_cpp/glm_one_group.cpp",
        "inmoose/edgepy/edgepy_cpp/interpolator.cpp",
        "inmoose/edgepy/edgepy_cpp/nbdev.cpp",
        "inmoose/edgepy/edgepy_cpp/objects.cpp",
    ],
    include_dirs=[numpy.get_include(), "inmoose/common_cpp/"],
    extra_compile_args=cxx_compile_flags,
    define_macros=macros,
)


setup(
    packages=[
        "inmoose",
        "inmoose/consensus_clustering",
        "inmoose/data",
        "inmoose/data/airway",
        "inmoose/data/pasilla",
        "inmoose/pycombat",
        "inmoose/common_cpp",
        "inmoose/edgepy",
        "inmoose/edgepy/edgepy_cpp",
        "inmoose/sim",
        "inmoose/utils",
        "inmoose/deseq2",
    ],
    package_data={
        "inmoose/edgepy/edgepy_cpp": [
            "edgepy_cpp.pyx",
            "__init__.pxd",
            "add_prior.h",
            "add_prior_count.h",
            "ave_log_cpm.cpp",
            "compute_apl.cpp",
            "compute_nbdev.cpp",
            "edgepy_cpp.h",
            "fit_levenberg.h",
            "fit_one_group.cpp",
            "get_one_way_fitted.cpp",
            "glm.h",
            "initialize_levenberg.cpp",
            "initialize_levenberg.h",
            "interpolator.h",
            "maximize_interpolant.cpp",
            "objects.h",
            "utils.h",
        ],
        "inmoose/common_cpp": [
            "matrix.h",
        ],
        "inmoose/data/airway": [
            "airway.h5ad",
        ],
        "inmoose/data/pasilla": [
            "Dmel.BDGP5.25.62.DEXSeq.chr.gff",
            "geneIDsinsubset.txt",
            "pasilla_gene_counts.tsv",
            "pasilla_sample_annotation.csv",
            "treated1fb.txt",
            "treated2fb.txt",
            "treated3fb.txt",
            "untreated1fb.txt",
            "untreated2fb.txt",
            "untreated3fb.txt",
            "untreated4fb.txt",
        ],
    },
    ext_modules=[common_cpp, edgepy_cpp, stats_cpp],
)



import unittest
import numpy as np
import patsy
import scipy.stats

from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq
from inmoose.deseq2.fitNbinomGLMs import fitNbinomGLMs


class Test(unittest.TestCase):
    def test_optim(self):
        """test that optim gives same results"""
        dds = makeExampleDESeqDataSet(n=100, interceptMean=10, interceptSD=3, seed=51)
        dds = dds.estimateSizeFactors()
        dds = dds.estimateDispersions()
        # make a larger predictor to test scaling
        dds.obs["condition"] = scipy.stats.norm.rvs(
            loc=0, scale=1000, size=dds.n_obs, random_state=51
        )
        modelMatrix = patsy.dmatrix("~condition", dds.obs)
        fit = fitNbinomGLMs(
            dds,
            modelMatrix=modelMatrix,
            modelFormula="~condition",
            alpha_hat=dds.var["dispersion"],
            lambda_=(2, 2),
            renameCols=True,
            betaTol=1e-8,
            maxit=100,
            useOptim=True,
            useQR=True,
            forceOptim=False,
        )
        fitOptim = fitNbinomGLMs(
            dds,
            modelMatrix=modelMatrix,
            modelFormula="~condition",
            alpha_hat=dds.var["dispersion"],
            lambda_=(2, 2),
            renameCols=True,
            betaTol=1e-8,
            maxit=100,
            useOptim=True,
            useQR=True,
            forceOptim=True,
        )

        self.assertTrue(np.allclose(fit["betaMatrix"], fitOptim["betaMatrix"]))
        self.assertTrue(np.allclose(fit["betaSE"], fitOptim["betaSE"]))

        # test optim gives same lfcSE
        dds = makeExampleDESeqDataSet(n=100, m=10, seed=42)
        dds.X[:, 0] = [0, 0, 0, 0, 0, 1000, 1000, 0, 0, 0]
        dds = DESeq(dds, betaPrior=False)
        # beta iter = 100 implies optim used for fitting
        self.assertEqual(dds.var["betaIter"][0], 100)
        res1 = dds.results(contrast=["condition", "B", "A"])
        res2 = dds.results(contrast=[0, 1])
        self.assertTrue(np.allclose(res1.lfcSE, res2.lfcSE))
        self.assertTrue(np.allclose(res1.pvalue, res2.pvalue, equal_nan=True))



import unittest
import numpy as np
import pandas as pd
import patsy

from inmoose.deseq2 import (
    estimateSizeFactorsForMatrix,
    DESeqDataSet,
    makeExampleDESeqDataSet,
)


class Test(unittest.TestCase):
    def test_size_factor(self):
        """test that size factor works"""

        m = np.arange(1, 17).T.reshape(4, 4)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="geoMeans should be as long as the number of columns of counts",
        ):
            estimateSizeFactorsForMatrix(m, geoMeans=np.arange(1, 6))

        with self.assertRaisesRegex(
            ValueError,
            expected_regex="every gene contains at least one zero, cannot compute log geometric means",
        ):
            estimateSizeFactorsForMatrix(m, geoMeans=np.zeros(4))

        with self.assertRaisesRegex(IndexError, expected_regex=""):
            estimateSizeFactorsForMatrix(m, controlGenes="foo")

        sf = estimateSizeFactorsForMatrix(m, geoMeans=np.arange(1, 5))
        ref = [0.3495015, 0.9246941, 1.4964761, 2.0676788]
        self.assertTrue(np.allclose(sf, ref))
        sf = estimateSizeFactorsForMatrix(m, controlGenes=np.array([0, 1]))
        ref = [0.2520327, 0.9761184, 1.6906866, 2.4042385]
        self.assertTrue(np.allclose(sf, ref))

        # norm matrix works
        nm = m / np.exp(np.mean(np.log(m), 0))
        true_sf = np.array([2, 1, 1, 0.5])
        counts = 2 * m
        counts[0, :] *= 2
        counts[3, :] = m[3, :]
        dds = DESeqDataSet(counts, pd.DataFrame({"x": np.arange(1, 5)}), "~1")
        dds = dds.estimateSizeFactors(normMatrix=nm)
        self.assertTrue(np.allclose((dds.normalizationFactors / nm)[:, 0], true_sf))

        # make some counts with zeros
        true_sf = 2.0 ** np.array([-2, -2, -1, -1, 0, 0, 0, 0, 1, 1, 2, 2])

        def dmr(x):
            return 0.01

        dds = makeExampleDESeqDataSet(sizeFactors=true_sf, n=100, dispMeanRel=dmr)
        cts = dds.X
        # set one random zero per row
        rng = np.random.default_rng(1)
        idx = rng.choice(np.arange(cts.shape[1]), size=cts.shape[0], replace=True)
        for i in range(cts.shape[0]):
            cts[i, idx[i]] = 0
        cts[0, 0] = 1000000  # an outlier
        dds.X = cts

        # positive counts method
        dds = dds.estimateSizeFactors(type_="poscounts")
        sf = dds.sizeFactors
        coefs = np.linalg.lstsq(patsy.dmatrix("~true_sf"), sf, rcond=None)[0]
        self.assertLess(np.abs(coefs[0]), 0.1)
        self.assertLess(np.abs(coefs[1] - 1), 0.1)

        # iterate method
        # TODO not implemented yet
        # dds = dds.estimateSizeFactors(type_="iterate")
        # sf = dds.sizeFactors
        # coefs = np.linalg.lstsq(patsy.dmatrix("~true_sf"), sf, rcond=None)[0]
        # self.assertLess(np.abs(coefs[0]), .1)
        # self.assertLess(np.abs(coefs[1]-1), .1)



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/plot.R' of the Bioconductor DESeq2 package
# (version 3.16).


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def plotDispEsts_dds(
    self,
    ymin=None,
    CV=False,
    genecol="black",
    fitcol="red",
    finalcol="dodgerblue",
    legend=True,
    xlab=None,
    ylab=None,
    log="xy",
    cex=0.7,
):
    """
    A simple helper function that plots the per-gene dispersion estimates
    together with the fitted mean-dispersion relationship.

    Arguments
    ---------
    ymin
        the lower bound for points on the plot, points beyond this are drawn as
        triangles at :code:`ymin`
    CV : bool
        whether to plot the asymptotic or biological coefficient of variation
        (the square root of dispersion) on the y-axis. As the mean grows to
        infinity, the square root of dispersion gives the coefficient of
        variation for the counts. Default is :code:`False`, plotting
        dispersion.
    genecol : str
        the color for gene-wise dispersion estimates
    fitcol : str
        the color of the fitted estimates
    finalcol : str
        the color of the final estimates used for testing
    legend : bool
        whether to draw a legend
    xlab : str
    ylab : str
    log : {"", "x", "y", "xy"}
        the axis (if any) to log scale
    cex : float
        the default size of the points to plot
    """

    if xlab is None:
        xlab = "mean of normalized counts"
    if ylab is None:
        if CV:
            ylab = "coefficient of variation"
        else:
            ylab = "dispersion"

    px = self.var["baseMean"]
    sel = px > 0
    px = px[sel]

    # transformation of dispersion into CV or not
    if CV:
        f = np.sqrt
    else:
        f = lambda x: x

    py = f(self.var["dispGeneEst"][sel])
    if ymin is None:
        ymin = 10 ** np.floor(np.log10(np.nanmin(py[py > 0])) - 0.1)

    plt.scatter(px, np.maximum(py, ymin), s=cex, c=genecol, label="gene-est")
    if "x" in log:
        plt.xscale("log")
    if "y" in log:
        plt.yscale("log")

    outliers = (self.var["dispOutlier"] == 1.0)[sel]

    if self.dispersions is not None:
        plt.scatter(
            px[~outliers],
            f(self.dispersions[sel][~outliers]),
            c=finalcol,
            s=cex,
            label="final",
        )
        # use a circle over outliers
        plt.scatter(
            px[outliers],
            f(self.dispersions[sel][outliers]),
            s=8 * cex,
            marker="o",
            facecolors="none",
            edgecolors=finalcol,
        )

    if self.var["dispFit"] is not None:
        plt.scatter(px, f(self.var["dispFit"][sel]), c=fitcol, s=cex, label="fitted")

    if legend:
        plt.legend(loc="lower right")

    plt.xlabel(xlab)
    plt.ylabel(ylab)
    plt.show()


def plotMA_dds(
    self,
    alpha=0.1,
    main="",
    xlab="mean of normalized counts",
    ylim=None,
    MLE=False,
    *args,
    **kwargs,
):
    """
    Helper function to MA-plot the results of :code:`self`
    """
    res = self.results(*args, **kwargs)
    return res.plotMA(alpha=alpha, main=main, xlab=xlab, ylim=ylim, MLE=MLE)


def plotMA_res(
    self,
    alpha=None,
    main="",
    xlab="mean of normalized counts",
    ylab="log fold change",
    ylim=None,
    colNonSig="grey",
    colSig="blue",
    colLine="grey",
    returnData=False,
    MLE=False,
    cex=1,
    log="x",
):
    """
    MA-plot from base means and log fold changes

    A simple helper function that makes a so-called "MA-plot", *i.e.* a scatter
    plot of log2 fold changes (on the y-axis) versus the mean of normalized
    counts (on the x-axis).

    This function also contains the code of the :code:`plotMA` function from
    the :code:`geneplotter` package.

    If :code:`self` contains a column :code:`"svalue"` then these will be used
    for coloring the points (with a default :code:`alpha=0.005`).

    Arguments
    ---------
    alpha : float
        the significance level for thresholding adjusted *p*-values
    main : str, optional
        title for the plot
    xlab : str, optional
        x-axis label, defaults to "mean of normalized counts"
    ylim : pair of floats, optional
        y limits
    colNonSig : str
        color to use for non-significant data points
    colSig : str
        color to use for significant data points
    colLine : str
        color to use for the horizontal (y=0) line
    returnData : bool
        whether to return the DataFrame instead of plotting
    MLE : bool
        if :code:`betaPrior=True` was used, whether to plot the MLE (unshrunken
        estimates), defaults to :code:`False`.  Requires that
        :meth:`.DESeqDataSet.results` was run with :code:`addMLE=True`.  Note
        that the MLE will be plotted regardless of this argument, if
        :func:`DESeq` wasrun with :code:`betaPrior=False`. See
        :func:`lfcShrink` for examples on how to plot shrunken log2 fold
        changes.

    Returns
    -------
    matplotlib.pyplot.Axes
        the axes object to be plotted with :code:`matplotlib.pyplot.show()`
    """
    sval = "svalue" in self.columns

    if sval:
        test_col = "svalue"
    else:
        test_col = "padj"

    if MLE:
        if "lfcMLE" not in self.columns:
            raise ValueError(
                "lfcMLE column is not present: you should first run results() with addMLE=True"
            )
        lfc_col = "lfcMLE"
    else:
        lfc_col = "log2FoldChange"

    if alpha is None:
        if sval:
            alpha = 0.005
            log.info("thresholding s-values on alpha=0.005 to color points")
        else:
            if self.alpha is None:
                alpha = 0.1
            else:
                alpha = self.alpha

    isDE = np.where(np.isnan(self[test_col]), False, self[test_col] < alpha)
    df = pd.DataFrame({"mean": self["baseMean"], "lfc": self[lfc_col], "isDE": isDE})

    if returnData:
        return df

    # R code calls geneplotter::plotMA, which is directly ported in Python below
    df = df[df["mean"] != 0]
    py = df["lfc"]
    if ylim is None:
        ylim = np.array([-1, 1]) * np.percentile(np.abs(py[py.is_finite()]), 0.99) * 1.1

    # markers:
    #   - up triangle (6) if below ylim[0]
    #   - down triangle (2) if above ylim[1]
    #   - circle (16) otherwise
    py_lo = py < ylim[0]
    py_hi = py > ylim[1]
    py_in = ~(py_lo | py_hi)

    colors = np.where(df["isDE"], colSig, colNonSig)

    ax = plt.gca()
    ax.scatter(df["mean"][py_in], py[py_in], marker="o", c=colors[py_in], s=cex)
    ax.scatter(
        df["mean"][py_lo],
        np.repeat(ylim[0], py_lo.sum()),
        marker="v",
        c=colors[py_lo],
        s=cex,
    )
    ax.scatter(
        df["mean"][py_hi],
        np.repeat(ylim[1], py_hi.sum()),
        marker="^",
        c=colors[py_hi],
        s=cex,
    )
    ax.axhline(y=0, c=colLine)

    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)

    if "x" in log:
        ax.set_xscale("log")
    if "y" in log:
        ax.set_yscale("log")

    return ax



def pasilla():
    """
    Retrieve the pasilla dataset as an :class:`AnnData` object

    The pasilla dataset is an RNA-Seq experiment on the effect of RNAi
    knockdown of Pasilla, the Drosophila melanogaster ortholog of mammalian
    NOVA1 and NOVA2, on the transcriptome [Brooks2011]_.
    """
    import importlib.resources
    import pandas as pd
    import anndata as ad

    data_dir = importlib.resources.files("inmoose.data.pasilla")
    cts = pd.read_csv(
        data_dir.joinpath("pasilla_gene_counts.tsv"), sep="\t", index_col=0
    )
    anno = pd.read_csv(data_dir.joinpath("pasilla_sample_annotation.csv"), index_col=0)

    # The columns of `cts` and the rows of `anno` use different labels and are
    # not in the same order. We first need to harmonize them before building the
    # AnnData object.

    # first get rid of the "fb" suffix
    anno.index = [i[:-2] for i in anno.index]

    # second reorder the index
    anno = anno.reindex(cts.columns)

    # we are now ready to build the AnnData object
    adata = ad.AnnData(cts.T, anno)

    return adata



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import numpy as np
import pandas as pd

from ..utils import Factor


def checkFullRank(modelMatrix):
    if np.linalg.matrix_rank(modelMatrix) < modelMatrix.shape[1]:
        if np.apply_along_axis(lambda col: np.all(col == 0), 0, modelMatrix).any():
            raise ValueError(
                "the model matrix is not full rank, so the model cannot be fit as specified. Levels or combinations of levels without any samples have resulted in column(s) of zeros in the model matrix."
            )
        else:
            raise ValueError(
                "the model matrix is not full rank, so the model cannot be fit as specified. One or more variables or interaction terms in the design formula are linear combinations of the others and must be removed."
            )


def buildMatrixWithNACols(m, NACols):
    mFull = np.full((m.shape[0], len(NACols)), np.nan)
    mFull[:, ~NACols] = m
    return mFull


def buildVectorWithNACols(v, NACols):
    vFull = np.full(len(NACols), np.nan)
    vFull[~NACols] = v
    return vFull


def buildDataFrameWithNACols(d, NACols):
    v = buildMatrixWithNACols(d.values.T, NACols).T
    return pd.DataFrame(v, columns=d.columns)


def nOrMoreInCell(modelMatrix, n):
    """for each sample in the model matrix,
    are there n or more replicates in the same cell (including that sample)
    """
    cells = Factor([tuple(modelMatrix[i]) for i in range(modelMatrix.shape[0])])
    return (cells.value_counts() >= n)[cells]


def renameModelMatrixColumns(data, design):
    """convenience function to make more descriptive names for factor variables"""
    factors = [
        info
        for f, info in design.design_info.factor_infos.items()
        if info.type == "categorical"
    ]
    return {
        f"{f.factor.name()}[T.{lvl}]": f"{f.factor.name()}_{lvl}_vs_{f.categories[0]}"
        for f in factors
        for lvl in f.categories[1:]
    }


def getFactorName(name):
    """convenience function to extract the name of a categorical factor"""
    if name.startswith("C("):
        end = name.find(",")
        if end < 0:
            end = name.find(")")
        return name[2:end]
    return name


def cleanCategoricalColumnName(name):
    """convenience function to remove the "C(" part of a categorical factor in a column name"""
    bracket = name.find("[")
    if bracket >= 0:
        return getFactorName(name[:bracket]) + name[bracket:]
    else:
        return name



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/AllClasses.R' and 'R/methods.R' of the
# Bioconductor DESeq2 package (version 3.16).


from collections import OrderedDict
import logging
import numpy as np
import pandas as pd
import patsy
from anndata import AnnData
from scipy.stats import median_abs_deviation as mad
from scipy.stats import norm

from ..utils import Factor, rnbinom
from .misc import buildVectorWithNACols, checkFullRank, cleanCategoricalColumnName


class MetaDataBase:
    def __init__(self, obj, name):
        self.obj = obj
        self.name = name
        if self.name not in self.obj.attrs:
            self.obj.attrs[self.name] = {}

    def __getitem__(self, index):
        return self.obj.attrs[self.name].get(index)

    def __setitem__(self, index, value):
        self.obj.attrs[self.name][index] = value

    def __delitem__(self, index):
        del self.obj.attrs[self.name][index]

    def __repr__(self):
        return repr(self.obj.attrs[self.name])

    def filter(self, items=None, like=None, regex=None):
        import re

        if len([x for x in [items, like, regex] if x is not None]) != 1:
            raise ValueError("pass exactly one argument to filter")
        if items is not None:
            if isinstance(items, str):
                items = [items]
            cols = [k for k, v in self.obj.attrs[self.name].items() if v in items]
        if like is not None:
            cols = [k for k, v in self.obj.attrs[self.name].items() if like in v]
        if regex is not None:
            cols = [
                k
                for k, v in self.obj.attrs[self.name].items()
                if v is not None and re.search(regex, v)
            ]

        return self.obj.filter(cols)


@pd.api.extensions.register_dataframe_accessor("type")
class TypeMetaData(MetaDataBase):
    def __init__(self, obj):
        super().__init__(obj, "type")


@pd.api.extensions.register_dataframe_accessor("description")
class DescMetaData(MetaDataBase):
    def __init__(self, obj):
        super().__init__(obj, "description")


class DESeqDataSet(AnnData):
    """
    DESeqDataSet extends AnnData class to store observations (samples) of
    variables/features in the rows of a matrix. AnnData also offers a
    convenient way to store observation-wise (per row) and feature-wise (per
    column) data.

    See Also
    --------
    .makeExampleDESeqDataSet
    """

    from .dispersions import estimateDispersions_dds as estimateDispersions
    from .estimateSizeFactors import estimateSizeFactors_dds as estimateSizeFactors
    from .results import results_dds as results
    from .plot import plotDispEsts_dds as plotDispEsts

    def __init__(self, countData, clinicalData=None, design=None, ignoreRank=False):
        """
        DESeqDataSet constructor

        Arguments
        ---------
        countData : pandas.DataFrame or DESeqDataSet
            Raw counts (if pandas.DataFrame) or DESeqDataSet object to copy.
            Raw counts matrix has one column per feature (usually genes) and
            one row per observation (usually single cells or samples).
        clinicalData : pandas.DataFrame
            Observation-wise clinical data. Arbitrary number of columns, as many
            rows as in countData.
        design : patsy.DesignMatrix-like
            An object (formula, matrix...) representing the design matrix. It
            will be fed to patsy.dmatrix and the resulting patsy.DesignMatrix
            will be stored in the DESeqDataSet.
        ignoreRank : bool, optional
            Whether to ignore testing that the design matrix is full rank.
        """

        if isinstance(countData, AnnData):
            if any((clinicalData,)):
                raise ValueError(
                    "If `countData` is an AnnData, no further arguments are needed"
                )
        super().__init__(X=countData, obs=clinicalData, dtype=int)

        if design is not None:
            self.design = design
            if not ignoreRank:
                checkFullRank(self.design)

        if isinstance(countData, DESeqDataSet):
            for k, v in countData.__dict__.items():
                if k not in self.__dict__:
                    self.__dict__[k] = v
        else:
            self.modelMatrix = None
            self.modelMatrixType = None
            self.weightsOK = None

    def copy(self):
        """deep copy of self"""
        res = __class__(super().copy())
        res.design = self.design
        for k, v in self.__dict__.items():
            if k not in res.__dict__:
                res.__dict__[k] = v
        return res

    @property
    def design(self):
        """design matrix"""
        return self.obsm["design"]

    @design.setter
    def design(self, d):
        # if d is already a design matrix, then this is a no-op
        # if design is already a matrix, then it is normalized (if needed) and other metadata specific to a design matrix are added. The `data=self.obs` part is ignored.
        design = patsy.dmatrix(d, data=self.obs, NA_action="raise")

        # ensure all categorical factors are encapsulated in a "C"
        for f in design.design_info.factor_infos.values():
            if f.type == "categorical":
                factor_name = f.factor.name()
                if not factor_name.startswith("C("):
                    factor_name = patsy.EvalFactor(f"C({factor_name})").name()
                cat_data = patsy.categorical.C(f.factor.eval(f.state, self.obs)).data
                self.obs[factor_name] = pd.Categorical(cat_data)

        # clean up the column names of the design matrix
        design.design_info.column_name_indexes = OrderedDict(
            [
                (cleanCategoricalColumnName(n), v)
                for n, v in design.design_info.column_name_indexes.items()
            ]
        )

        self.obsm["design"] = design

    @property
    def dispersions(self):
        """
        accessor for the dispersions
        """
        try:
            return self.var["dispersion"]
        except KeyError:
            return None

    @dispersions.setter
    def dispersions(self, disp):
        self.var["dispersion"] = disp

    @dispersions.deleter
    def dispersions(self):
        del self.var["dispersion"]

    @property
    def sizeFactors(self):
        """
        accessor for the size factors

        The size factors assign to each sample of the count matrix a value, the
        size factor, such that sample-wise count values can be brought to a
        common scale by dividing by the corresponding size factor (as performed
        by :code:`dds.counts(normalized=True)`.  See :func:`.DESeq` for a
        description of the use of size factors. If gene-specific normalization
        is desired for each sample, use :attr:`normalizationFactors`.
        """
        try:
            return self.obs["sizeFactors"]
        except KeyError:
            return None

    @sizeFactors.setter
    def sizeFactors(self, sf):
        if not np.all(~np.isnan(sf)):
            raise ValueError("size factors should not be nan")
        if not np.all(np.isfinite(sf)):
            raise ValueError("size factors should be finite")
        if not np.all(np.greater(sf, 0)):
            raise ValueError("size factors should be positive")
        self.obs["sizeFactors"] = sf

    @sizeFactors.deleter
    def sizeFactors(self):
        del self.obs["sizeFactors"]

    @property
    def normalizationFactors(self):
        """
        accessor for the normalization factors

        Gene-specific normalization factors for eacg sample can be provided as
        a matrix, which will preempt :attr:`sizeFactors`. In some experiments,
        counts for each sample have varying dependence on covariates, *e.g.* on
        GC-content for sequencing data run on different days, and in this case
        it makes sense to provide gene-specific factors for each sample rather
        than a single size factor.

        Notes
        -----

        Normalization factors alter the model of :func:`.DESeq` in the
        following way, for counts :math:`K_{ij}` and normalization factors
        :math:`NF_{ij}` for gene :math:`i` and sample :math:`j`:

        .. math::
           K_{ij} \sim \mathcal{NB}(\mu_{ij}, \\alpha_i)
           \mu_{ij} = NF_{ij} q_{ij}

        Normalization factors are on the scale of the counts (similar to
        :attr:`sizeFactors`) and unlike offsets, which are typically on the
        scale of the predictors (in this case, log counts). Normalization
        factors should include library size normalization. They should have
        gene-wise geometric mean near 1, as is the case with size factors, such
        that the mean of normalized counts is close to the mean of unnormalized
        counts.
        """
        try:
            return self.layers["normalizationFactors"]
        except KeyError:
            return None

    @normalizationFactors.setter
    def normalizationFactors(self, nf):
        if not np.all(~np.isnan(nf)):
            raise ValueError("normalization factors should not be nan")
        if not np.all(np.isfinite(nf)):
            raise ValueError("normalization factors should be finite")
        if not np.all(np.greater(nf, 0)):
            raise ValueError("normalization factors should be positive")
        self.layers["normalizationFactors"] = nf

    @normalizationFactors.deleter
    def normalizationFactors(self):
        del self.layers["normalizationFactors"]

    @property
    def dispersionFunction(self):
        """
        dispersion function

        The dispersion function is calculated by
        :meth:`DESeqDataSet.estimateDispersions` and used by
        :func:`varianceStabilizingTransformation`. Parametric dispersion fits
        store the coefficients of the fit as attributes in this property.

        This property cannot be set directly, only through
        :meth:`DESeqDataSet.setDispFunction`. Setting it will also overwrite
        :code:`self.var["dispFit"]` and the estimate of the variance of
        dispersion residuals.
        """
        return self._dispersionFunction

    @dispersionFunction.setter
    def dispersionFunction(self, v):
        raise RuntimeError(
            "do not write dispersionFunction, use setDispFunction method"
        )

    def setDispFunction(self, value, estimateVar=True):
        """
        Setter for the :attr:`DESeqDataSet.dispersionFunction` property

        This function also overwrites :code:`self.var["dispFit"]` and the
        estimate of the variance of dispersion residuals (see argument
        :code:`estimateVar`).

        See also
        --------
        estimateDispersions

        Arguments
        ---------
        value : function
            a function
        estimateVar : bool
            whether to estimate the variance of the dispersion residuals.
            Setting this to :code:`False` is needed, *e.g.* in
            :func:`.estimateDispersionsMAP`, when called on a subset of the full
            dataset in parallel execution.

        Returns
        -------
        DESeqDataSet
            self object
        """
        # the following will add 'dispFit' to self.var
        # first check to see that we have 'baseMean' and 'allZero'
        if "baseMean" not in self.var or "allZero" not in self.var:
            self = self.getBaseMeansAndVariances()
        # warning about existing 'dispFit' data will be removed
        if "dispFit" in self.var:
            del self.var["dispFit"]
        # now call the dispersionFunction on 'baseMean' to make 'dispFit'
        nonzeroIdx = ~self.var["allZero"]
        dispFit = value(self.var["baseMean"][nonzeroIdx])
        self.var["dispFit"] = buildVectorWithNACols(dispFit, self.var["allZero"])
        self.var.type["dispFit"] = "intermediate"
        self.var.description["dispFit"] = "fitted values of dispersion"

        # estimate variance of log dispersion around the fit
        if estimateVar:
            # need to estimate variance of log dispersion residuals
            minDisp = 1e-8
            dispGeneEst = self.var["dispGeneEst"][nonzeroIdx]
            aboveMinDisp = dispGeneEst >= minDisp * 100
            if np.nansum(aboveMinDisp) > 0:
                dispResiduals = np.log(dispGeneEst) - np.log(dispFit)
                varLogDispEsts = (
                    mad(dispResiduals[aboveMinDisp], nan_policy="omit") ** 2
                )
                value.varLogDispEsts = varLogDispEsts
            else:
                logging.info(
                    "variance of dispersion residuals not estimated (necessary only for differential expression calling)"
                )

        # store the dispersion function
        self._dispersionFunction = value
        return self

    def __getitem__(self, index):
        res = self.__class__(super().__getitem__(index).copy())
        # make sure the design is valid
        res.design = self.design.design_info
        # set the attributes already set in __init__
        res.modelMatrix = self.modelMatrix
        res.modelMatrixType = self.modelMatrixType
        res.weightsOK = self.weightsOK
        # other attributes
        for k, v in self.__dict__.items():
            if k not in res.__dict__:
                res.__dict__[k] = v
        # preserve metadata
        for c in self.var.columns:
            t = self.var.type[c]
            if t is not None:
                res.var.type[c] = t
            d = self.var.description[c]
            if d is not None:
                res.var.description[c] = d
        return res

    def counts(self, normalized=False, replaced=False):
        """
        Accessor for the counts of a :class:`DESeqDataSet`, which are stored as
        a matrix of non-negative integer, with one row per observation and one
        column per feature.

        Arguments
        ---------
        normalized : bool, optional
            Flag indicating whether to divide the counts by the size factors or
            normalization factors before returning. Normalization factors (if
            present) always preempt size factors.
        replaced : bool, optional
            Flag indicating whether to return the counts with outliers replaced
            instead of the original counts. This flag is meaningful only after
            a call to :func:`.DESeq`.

        Returns
        -------
        ndarray
            the count data, possibly normalized, possibly with outliers
            replaced
        """
        if replaced:
            if "replaceCounts" in self.layers:
                cnts = self.layers["replaceCounts"]
            else:
                logging.warnings.warn(
                    "there is no layer named 'replacedCounts', using original. calling DESeq() will replace outliers if they are detected and store this layer."
                )
                cnts = self.X
        else:
            cnts = self.X

        if not normalized:
            return cnts
        else:
            if self.normalizationFactors is not None:
                return cnts / self.normalizationFactors
            elif self.sizeFactors is None or np.isnan(self.sizeFactors).any():
                raise ValueError(
                    "first calculate size factors, add normalizationFactors, or set normalized=False"
                )
            else:
                return cnts / self.sizeFactors.values[:, None]

    def designAndArgChecker(self, betaPrior):
        """
        Arguments
        ---------
        betaPrior : bool
        """
        di = self.design.design_info
        termsOrder = np.array([len(t.factors) for t in di.terms])
        hasIntercept = 0 in termsOrder
        interactionPresent = (termsOrder > 1).any()

        if betaPrior and not hasIntercept:
            raise ValueError(
                "betaPrior=True can only be used if the design has an intercept. If not, use betaPrior=False"
            )
        if betaPrior and interactionPresent:
            raise ValueError(
                "betaPrior=False should be used for designs with interactions"
            )

        if not betaPrior:
            if np.linalg.matrix_rank(self.design) < self.design.shape[1]:
                raise ValueError("full model matrix is not full rank")

    def getBaseMeansAndVariances(self):
        """Get base means and variances

        An internally used function to calculate the gene (columns) means
        and variances from the normalized counts, which requires that
        estimateSizeFactors has already been called. Adds these and a Boolean
        var to identify the columns whose sum is zero.
        """
        cts_norm = self.counts(normalized=True)
        if "weights" in self.layers:
            wts = self.layers["weights"]
            cts_norm = wts * cts_norm
        self.var["baseMean"] = np.mean(cts_norm, 0)
        self.var.type["baseMean"] = "intermediate"
        self.var.description["baseMean"] = "mean of normalized counts for all samples"
        self.var["baseVar"] = np.var(cts_norm, 0, ddof=1)
        self.var.type["baseVar"] = "intermediate"
        self.var.description[
            "baseVar"
        ] = "variance of normalized counts for all samples"
        self.var["allZero"] = np.sum(self.counts(), 0) == 0
        self.var.type["allZero"] = "intermediate"
        self.var.description["allZero"] = "all counts for a gene are zero"

        return self

    def getSizeOrNormFactors(self):
        """simple function to return a matrix of size factors or normalization factors"""
        if self.normalizationFactors is not None:
            return self.normalizationFactors
        else:
            return self.sizeFactors.to_numpy()[:, None].repeat(self.n_vars, axis=1)

    def resultsNames(self):
        """
        Return the names of the estimated effects (coefficients) of the model

        Returns
        -------
        index
            the names of the columns available as results, usually a
            combination of the variable name and a level
        """
        return self.var.description.filter(regex="log2 fold change").columns

    def removeResults(self):
        """
        remove results columns from the object

        Returns
        -------
        DESeqDataSet
            :code:`self` with results data removed
        """
        self.var = self.var.drop(self.var.type.filter("results").columns, axis=1)
        return self

    def makeExpandedModelMatrix(self):
        data = self.obs.apply(
            lambda col: Factor(col)
            .add_categories(["__null__"])
            .reorder_categories(col.dtype.categories.insert(0, "__null__"))
            if isinstance(col.dtype, pd.CategoricalDtype)
            else col
        )
        # formula = "+".join([f"C({t.name()}, Treatment(reference='__null__'))"
        formula = "+".join(
            [
                f"{t.name()}" if len(t.factors) != 0 else "1"
                for t in self.design.design_info.terms
            ]
        )
        return patsy.dmatrix(formula, data=data)

    def getDesignFactors(self):
        return [
            f.name()
            for f, info in self.design.design_info.factor_infos.items()
            if info.type == "categorical"
        ]

    def addAllContrasts(self, betaMatrix):
        """add all first order contrasts"""
        designFactors = self.getDesignFactors()
        coldata = self.obs
        for f in designFactors:
            lvls = coldata[f].dtype.categories
            mmColnames = [f"{f}[T.{c}]" for c in lvls]
            M = betaMatrix.filter(mmColnames)
            n = M.shape[1]
            if n > 1:
                ii = [f"{c}" for k in range(1, n) for c in M.columns[k:]]
                jj = [f"{c}" for k, c in enumerate(M.columns) for _ in range(n - k - 1)]
                contrastCols = pd.DataFrame(
                    {
                        f"{f}Cntrst{k}": M[i] - M[j]
                        for k, (i, j) in enumerate(zip(ii, jj))
                    }
                )
                betaMatrix = pd.concat([betaMatrix, contrastCols], axis=1)
        return betaMatrix

    def averagePriorsOverLevels(self, betaPriorVar):
        expandedModelMatrix = self.makeExpandedModelMatrix()
        expandedNames = expandedModelMatrix.design_info.column_names
        betaPriorIn = betaPriorVar
        betaPriorOut = pd.Series(np.zeros(len(expandedNames)), index=expandedNames)
        commonIndex = [n for n in expandedNames if n in betaPriorIn.columns]
        betaPriorOut[commonIndex] = betaPriorIn[commonIndex].values.squeeze()
        designFactors = self.getDesignFactors()
        for f in designFactors:
            lvls = self.obs[f].dtype.categories
            mmColnames = pd.Index([f"{f}[T.{c}]" for c in lvls]).append(
                betaPriorIn.filter(regex=f"^{f}Cntrst\d+$").index
            )
            meanPriorVar = np.mean(betaPriorIn.filter(mmColnames).values)
            betaPriorOut[np.isin(betaPriorOut.index, mmColnames)] = meanPriorVar

        if np.any(np.isnan(betaPriorOut)):
            raise ValueError("beta prior is NA for some cols")
        if not np.all(betaPriorOut > 0):
            raise ValueError("beta prior is <= 0 for some cols")

        return betaPriorOut


def makeExampleDESeqDataSet(
    n=1000,
    m=12,
    betaSD=0,
    interceptMean=4,
    interceptSD=2,
    dispMeanRel=lambda x: 4 / x + 1,
    sizeFactors=None,
    seed=None,
):
    """
    Make a simulated :class:`.DESeqDataSet`

    This function constructs a simulated dataset of Negative Binomial data
    from two conditions. By default, there are no fold changes between the two
    conditions, but this can be adjusted with the :code:`betaSD` argument.

    See also
    --------
    ~inmoose.sim.sim_rnaseq
        another function to create simulated Negative Binomial data, accounting
        for conditions, outliers and batches

    Arguments
    ---------
    n : int
        the number of genes
    m : int
        the number of samples
    betaSD : float
        the standard deviation for non-intercept betas, i.e.
        :math:`\\beta \sim \mathcal{N}(0, betaSD)`
    interceptMean : float
        the mean of the intercept betas (log2 scale)
    interceptSD : float
        the standard deviation of the intercept betas (log2 scale)
    dispMeanRel
        a function specifying the relationship of the dispersions on the
        :code:`2^trueIntercept`
    sizeFactors : array-like
        multiplicative factors for each sample

    Returns
    -------
    DESeqDataSet
        a :class:`.DESeqDataSet` with true dispersion, intercept and beta values
        in :attr:`DESeqDataSet.var`. Note that the true betas are provided on
        the log2 scale.
    """
    if sizeFactors is None:
        sizeFactors = np.ones(m)
    sizeFactors = np.asarray(sizeFactors)

    rng = np.random.default_rng(seed)
    beta = np.column_stack(
        [
            norm(loc=interceptMean, scale=interceptSD).rvs(n, random_state=rng),
            norm(loc=0, scale=betaSD).rvs(n, random_state=rng),
        ]
    )
    dispersion = dispMeanRel(2 ** (beta[:, 0]))
    clinData = pd.DataFrame(
        {
            "condition": Factor(
                np.concatenate(
                    [
                        np.full(int(np.ceil(m / 2)), "A"),
                        np.full(int(np.floor(m / 2)), "B"),
                    ]
                )
            )
        }
    )
    clinData.index = [f"sample{i}" for i in range(m)]

    if m > 1:
        x = patsy.dmatrix("~clinData['condition']")
    else:
        x = np.column_stack([np.ones(m, dtype=int), np.zeros(m, dtype=int)])

    mu = 2.0 ** (x @ beta.T) * sizeFactors.reshape(x.shape[0], 1)
    countData = rnbinom((m, n), mu=mu, size=1 / dispersion, seed=rng)
    countData = pd.DataFrame(countData, index=[f"sample{i}" for i in range(m)])

    if m > 1:
        design = "~condition"
    else:
        design = "~1"

    obj = DESeqDataSet(countData=countData, clinicalData=clinData, design=design)
    obj.var["trueIntercept"] = beta[:, 0]
    obj.var.type["trueIntercept"] = "input"
    obj.var.description["trueIntercept"] = "simulated intercept values"
    obj.var["trueBeta"] = beta[:, 1]
    obj.var.type["trueBeta"] = "input"
    obj.var.description["trueBeta"] = "simulated beta values"
    obj.var["trueDisp"] = dispersion
    obj.var.type["trueDisp"] = "input"
    obj.var.description["trueDisp"] = "simulated dispersion values"

    return obj
